from typing import Any
import fs

from . import base
from .compiler_base import (
    CompilerFloatBehavior, CompilerWarningState, LanguageStandard, CompilerWarningSettings, CompilerWarningLevel,
    UnsupportedSettingError, CompilerArgFormatter, LanguageConformance, CompilerSettings, UnsupportedLanguageError,
    CompilerOptimizationMode
)


class CompilerArgFormatterMSVC(CompilerArgFormatter):
    def format_beginning_args(self) -> list[str]:
        return ["/nologo"]

    def format_object_compilation_args(self) -> list[str]:
        return [
            "/c",
            "\"%1\"",
            "/Fo\"%2\"",
        ]

    def format_debug_info_args(self, output_dir_path: str, output_base_name: str) -> list[str]:
        intermediate_pdb_path = fs.path.join(output_dir_path, f"{output_base_name}-lib.pdb")
        return [
            f"/Fd\"{intermediate_pdb_path}\"",
        ]

    def format_pch_creation_args(self, pch_header_path: str, pch_output_path: str) -> list[str]:
        return [
            "/c",
            "\"%1\"",  # PCH input file (.cpp)
            "/Fo\"%3\"",  # PCH output library (.lib)
            f"/Yc\"{fs.path.basename(pch_header_path)}\"",
            "/Fp\"%2\""  # PCH output file (compiled .pch)
        ]

    def format_pch_usage_args(self, pch_header_path: str, pch_output_path: str) -> list[str]:
        return [
            f"/Yu\"{fs.path.basename(pch_header_path)}\"",
            f"/Fp\"{pch_output_path}\"",
        ]

    def format_definition_compiler_arg(self, name: str, value: base.Any) -> str:
        if value is not None:
            return f"/D\"{name}={value}\""
        else:
            return f"/D\"{name}\""

    def format_include_dir_arg(self, include_dir: str) -> str:
        return f"/I\"{include_dir}\""

    def format_system_include_dir_arg(self, include_dir: str) -> str:
        return f"/I\"{include_dir}\""

    @staticmethod
    def _format_language_std(std: LanguageStandard) -> list[str]:
        match std:
            case LanguageStandard.C89:
                raise UnsupportedLanguageError("C language standards older than C11 are not supported.")
            case LanguageStandard.C99:
                raise UnsupportedLanguageError("C language standards older than C11 are not supported.")
            case LanguageStandard.C11:
                return ['/TC', '/std:c11']
            case LanguageStandard.C17:
                return ['/TC', '/std:c17']
            case LanguageStandard.C23:
                return ['/TC', '/std:c23']
            case LanguageStandard.C_LATEST:
                return ['/TC', '/std:clatest']
            case LanguageStandard.CPP03:
                raise UnsupportedLanguageError("C++ language standards older than C++14 are not supported.")
            case LanguageStandard.CPP11:
                raise UnsupportedLanguageError("C++ language standards older than C++14 are not supported.")
            case LanguageStandard.CPP14:
                return ['/TP', '/std:c++14']
            case LanguageStandard.CPP17:
                return ['/TP', '/std:c++17']
            case LanguageStandard.CPP20:
                return ['/TP', '/std:c++20']
            case LanguageStandard.CPP23:
                return ['/TP', '/std:c++23']
            case LanguageStandard.CPP_LATEST:
                return ['/TP', '/std:c++latest']
            case _:
                raise UnsupportedLanguageError(f"Unhandled language standard '{std}'.")

    @staticmethod
    def _format_language_conformance(lang_conformance: LanguageConformance) -> list[str]:
        match lang_conformance:
            case LanguageConformance.MIN:
                return ["/permissive", "/volatile:ms"]
            case LanguageConformance.MAX:
                return ["/permissive-", "/volatile:iso"]
            case _:
                raise UnsupportedLanguageError(f"Unhandled language conformance option '{lang_conformance}'.")

    @staticmethod
    def _format_warning_level(warning_level: CompilerWarningLevel) -> list[str]:
        match warning_level:
            case CompilerWarningLevel.W0:
                return ["/W0"]
            case CompilerWarningLevel.W1:
                return ["/W1"]
            case CompilerWarningLevel.W2:
                return ["/W2"]
            case CompilerWarningLevel.W3:
                return ["/W3"]
            case CompilerWarningLevel.W4:
                return ["/W4"]
            case CompilerWarningLevel.All:
                return ["/Wall"]
            case _:
                raise UnsupportedSettingError()

    @staticmethod
    def _format_warning_version(ver: str) -> list[str]:
        return [f"/Wv:{ver}"]

    @staticmethod
    def _format_warning_state(warn: str, state: CompilerWarningState) -> list[str]:
        match state:
            case CompilerWarningState.DISABLE:
                return [f"/wd{warn}"]
            case CompilerWarningState.ENABLE:
                return [f"/w1{warn}"]
            case CompilerWarningState.REPORT_ONCE:
                return [f"/wo{warn}"]
            case CompilerWarningState.TREAT_AS_ERROR:
                return [f"/we{warn}"]
            case _:
                raise UnsupportedSettingError()

    @classmethod
    def _is_known_warning(cls, warn: str):
        return warn.isdigit()

    @classmethod
    def _format_warning_states(cls, states: dict[str, CompilerWarningState]) -> list[str]:
        l = []
        for warn, state in states.items():
            if cls._is_known_warning(warn):
                l += cls._format_warning_state(warn, state)
        return l

    @classmethod
    def _format_warning_settings(cls, warning_settings: CompilerWarningSettings) -> list[str]:
        l = []
        l += cls._format_warning_level(warning_settings.level)
        l += cls._format_warning_version(warning_settings.version)
        if warning_settings.treat_as_errors:
            l += ["/WX"]
        l += cls._format_warning_states(warning_settings.warning_states)
        l += ['/experimental:external', '/external:anglebrackets', '/external:W0']
        return l

    @staticmethod
    def _format_float_behavior(beh: CompilerFloatBehavior) -> list[str]:
        match beh:
            case CompilerFloatBehavior.PRECISE:
                return ["/fp:precise"]
            case CompilerFloatBehavior.STRICT:
                return ["/fp:strict"]
            case CompilerFloatBehavior.FAST:
                return ["/fp:fast"]
            case _:
                raise UnsupportedSettingError()

    @staticmethod
    def _format_optimization_mode(opt: CompilerOptimizationMode) -> list[str]:
        match opt:
            case CompilerOptimizationMode.NONE:
                return ["/Od"]
            case CompilerOptimizationMode.MINIMIZE_SIZE:
                return ["/O1"]
            case CompilerOptimizationMode.MAXIMIZE_SPEED:
                return ["/O2"]

    @classmethod
    def _format_compiler_settings(cls, settings: CompilerSettings) -> list[str]:
        l = []
        # emit error for unknown arguments
        l += ["/options:strict"]
        l += cls._format_language_std(settings.lang_std)
        l += cls._format_language_conformance(settings.lang_conformance)
        l += cls._format_warning_settings(settings.warning_settings)
        l += ["/GR"] if settings.enable_rtti else ["/GR-"]
        l += ["/EHsc"] if settings.enable_exceptions else ["/EHs-"]
        l += cls._format_float_behavior(settings.float_behavior)
        # require to always define the class before declaring a pointer-to-member
        l += ["/vmb"]
        l += ["/MDd"] if settings.use_debug_runtime else ["/MD"]
        if settings.enable_debug_info:
            # debug format: PDB
            l += ["/Zi"]
            # faster PDB generation
            l += ["/Zf"]
        # force synchronous PDB writes, since fastbuild uses multiple cl.exe instances.
        l += ["/FS"]
        l += cls._format_optimization_mode(settings.optimization_mode)
        if settings.enable_security_checks:
            # enable buffer security checks
            l += ['/GS']
            l += ['/sdl']
            # enable control flow guards
            l += ['/guard:cf']
            # enable EH continuation metadata(must be also present in linker args)
            l += ['/guard:ehcont']
            # enable all runtime checks.RTCc rejects conformant code, so it is not supported by
            # the C++ Standard Library
            l += ['/RTCsu']
        else:
            # disable buffer security checks
            l += ['/GS-']
            # + '/sdl- ' // it overrides / GS -
            # disable control flow guards
            l += ['/guard:cf-']
            # disable EH continuation metadata
            l += ['/guard:ehcont-']
        return l

    def format_settings_args(self, settings: CompilerSettings) -> list[str]:
        return self._format_compiler_settings(settings)


class CompilerArgFormatterClangCL(CompilerArgFormatterMSVC):
    pass



from . import base
from .workspace_def import TargetKind
from .linker_base import LinkerArgFormatter, LinkerSettings, LinkerWarningState, LinkerWarningSettings


class LinkerArgFormatterLDLLD(LinkerArgFormatter):
    def format_input_output_declaration(
            self, output_dir_path: str, output_base_name: str, target_kind: TargetKind
    ) -> list[str]:
        args = ["\"%1\"", "-o \"%2\""]
        if target_kind == TargetKind.DYNAMIC_LIB:
            args += ["--shared"]
        return args

    def format_lib_arg(self, lib: str) -> str:
        return f"-l\"{lib}\""

    def format_lib_dir_arg(self, lib_dir: str) -> str:
        return f"-L\"{lib_dir}\""

    @staticmethod
    def _format_warning_state(warn: str, state: LinkerWarningState) -> list[str]:
        match state:
            case LinkerWarningState.ENABLE:
                return []
            case LinkerWarningState.DISABLE:
                return [f"/IGNORE:{warn}"]
            case LinkerWarningState.TREAT_AS_ERROR:
                return [f"/WX:{warn}"]
            case _:
                raise NotImplementedError()

    @classmethod
    def _format_warning_states(cls, states: dict[str, LinkerWarningState]) -> list[str]:
        l = []
        for warn, state in states.items():
            l += cls._format_warning_state(warn, state)
        return l

    @classmethod
    def _format_warning_settings(cls, settings: LinkerWarningSettings) -> list[str]:
        l = []
        l += ["/WX"] if settings.treat_as_errors else ["/WX:NO"]
        l += ["/MACHINE:X64"]
        l += cls._format_warning_states(settings.warning_states)
        return l

    def format_settings_args(self, settings: LinkerSettings) -> list[str]:
        l = []
        l += ["/NOLOGO"]
        l += self._format_warning_settings(settings.warning_settings)
        l += ["/DEBUG:FULL"] if settings.enable_debug_info else ["/DEBUG:NONE"]
        l += ["/GUARD:CF"] if settings.enable_security_checks else ["/GUARD:NO"]
        l += ["/OPT:REF,ICF,LBR"] if settings.enable_optimizations else ["/OPT:NOREF,NOICF,NOLBR"]
        return l


class LinkerArgFormatterLD(LinkerArgFormatterLDLLD):
    pass




from .librarian_base import LibrarianArgFormatter


class LibrarianArgFormatterLLVM(LibrarianArgFormatter):
    def format_input_output_args(self) -> list[str]:
        return ["rc", "%2", "%1"]


class LibrarianArgFormatterGCC(LibrarianArgFormatterLLVM):
    pass




from abc import ABC, abstractmethod
import sys
import subprocess
import os.path
from enum import Enum, auto
from typing import Any
from dataclasses import dataclass, field

from .compiler_base import (
    CompilerSettings
)

from .linker_base import (
    LinkerSettings
)


def append_unique(list_: list[Any], item: any):
    if item not in list_:
        list_.append(item)


def extend_unique(list_a: list[Any], list_b: list[Any]):
    for item in list_b:
        if item not in list_a:
            list_a.append(item)


def build_target_key(target_name: str, toolchain_name: str, build_type: str):
    return "{}_{}_{}".format(target_name, toolchain_name, build_type)


BUILD_TYPE_MATCH_ALL = "*"
BUILD_TYPE_DEBUG = "debug"
BUILD_TYPE_RELEASE = "release"
BUILD_TYPE_PROFILING = "profiling"

TOOLCHAIN_DEFAULT = "default"
TOOLCHAIN_VS16_MSVC = "vs16_msvc"
TOOLCHAIN_VS16_LLVM13 = "vs16_llvm13"
TOOLCHAIN_VS17_MSVC = "vs17_msvc"
TOOLCHAIN_VS17_LLVM13 = "vs17_llvm13"

OS_WINDOWS = "windows"


class Logger(ABC):
    @abstractmethod
    def log_info(self, *args):
        pass

    @abstractmethod
    def log_error(self, *args):
        pass


class LoggerDefault(Logger):
    def log_info(self, *args):
        print(*args, file=sys.stdout)

    def log_error(self, *args):
        print(*args, file=sys.stderr)


class ProcessRunner(ABC):
    @abstractmethod
    def run(self, args, *, env, check):
        pass


class ProcessRunnerDefault(ProcessRunner):
    def run(self, args, *, env, check):
        subprocess.run(args, env=env, check=check)


class ThirdPartyInvalidFormat(RuntimeError):
    pass


@dataclass
class ThirdPartyTargetInterface:
    include_dirs: list[str] = field(default_factory=list)
    definitions: dict[str, Any] = field(default_factory=list)
    link_libs: list[str] = field(default_factory=list)
    link_libs_dirs: list[str] = field(default_factory=list)
    load_time_libs: list[str] = field(default_factory=list)

    @staticmethod
    def from_dict(d: dict):
        if not isinstance(d, dict):
            raise ThirdPartyInvalidFormat()

        if not set(d.keys()).issubset({"include_dirs", "definitions", "link_libs", "link_libs_dirs", "load_time_libs"}):
            raise ThirdPartyInvalidFormat()

        include_dirs = d.get("include_dirs", [])
        if not isinstance(include_dirs, list):
            raise ThirdPartyInvalidFormat()

        definitions = d.get("definitions", {})
        if not isinstance(definitions, dict):
            raise ThirdPartyInvalidFormat()

        link_libs = d.get("link_libs", [])
        if not isinstance(link_libs, list):
            raise ThirdPartyInvalidFormat()

        link_libs_dirs = d.get("link_libs_dirs", [])
        if not isinstance(link_libs_dirs, list):
            raise ThirdPartyInvalidFormat()

        load_time_libs = d.get("load_time_libs", [])
        if not isinstance(load_time_libs, list):
            raise ThirdPartyInvalidFormat()

        return ThirdPartyTargetInterface(include_dirs=include_dirs, definitions=definitions, link_libs=link_libs,
                                         link_libs_dirs=link_libs_dirs, load_time_libs=load_time_libs)


class SystemConfigManifestProvider(ABC):
    @abstractmethod
    def run(self, working_dir_abs_path, build_types):
        pass

    @abstractmethod
    def get_toolchains_manifest(self):
        pass

    @abstractmethod
    def get_third_party_manifest(self):
        pass


class ToolchainSettings(ABC):
    @abstractmethod
    def get_build_types(self) -> set[str]:
        pass

    @abstractmethod
    def get_compiler_settings(self, build_type: str) -> CompilerSettings:
        pass

    @abstractmethod
    def get_compiler_definitions(self, build_type: str) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_linker_settings(self, build_type: str) -> LinkerSettings:
        pass


class ToolchainSettingsProvider(ABC):
    @abstractmethod
    def get_toolchain_settings(self, toolchain_name) -> ToolchainSettings:
        pass

    @abstractmethod
    def get_toolchains(self) -> set[str]:
        pass


class BuildScriptEmitter(ABC):
    @abstractmethod
    def filename(self):
        pass

    @abstractmethod
    def contents(self, wks_name, source_dir_abs_path, configure_dir_abs_path, build_dir_abs_path,
                 toolchains_manifest, toolchain_settings_provider, build_types, targets_names, targets_impls) -> str:
        pass



from __future__ import annotations

from typing import Any
import importlib.resources
import fs
import os.path
import json

from . import base
from . import conan_generator


class SystemConfigManifestProviderConan(base.SystemConfigManifestProvider):
    TOOLCHAIN_MANIFEST_FILENAME = "toolchain_gen.json"
    THIRD_PARTY_MANIFEST_FILENAME = "third_party_gen.json"

    IMPORTS_BIN_DIRNAME = "imports_bin"
    IMPORTS_LIB_DIRNAME = "imports_lib"

    SUPPORTED_CONAN_BUILD_TYPES = [
        "Debug",
        "Release"
    ]

    def __init__(self, *, logger, process_runner, conanfile_path, execute_conan, base_to_conan_build_type):
        self.logger = logger
        self.process_runner = process_runner
        self.conanfile_path = conanfile_path
        self.execute_conan = execute_conan
        self.base_to_conan_build_type = base_to_conan_build_type

        for build_type, conan_build_type in self.base_to_conan_build_type.items():
            if conan_build_type not in self.SUPPORTED_CONAN_BUILD_TYPES:
                raise RuntimeError(f"Unsupported conan build type '{conan_build_type}'")

        self.third_party_manifest: dict[str, Any] = {}
        self.toolchains_manifest: dict[str, Any] = {}

    @staticmethod
    def get_install_dir_name(build_type: str) -> str:
        return f"install-{build_type}"

    def run(self, working_dir_abs_path, build_types):

        for build_type in build_types:
            if build_type not in self.base_to_conan_build_type.keys():
                raise RuntimeError(f"Unsupported build type '{build_type}'")

        conan_export_invocation = []
        conan_export_resource = ""

        for resource in importlib.resources.contents(conan_generator):
            if resource == "conanfile.py":
                with importlib.resources.path(conan_generator, resource) as conan_generator_path:
                    conan_export_resource = resource
                    conan_export_invocation = ["conan", "export", f"{conan_generator_path}"]
                break

        if conan_export_resource == "":
            raise RuntimeError("Internal error: unable to find Conan generator resource.")

        conan_install_invocations = []
        conan_install_build_types = []

        # Run toolchain discovery process (in the generator) only for a single build type.
        toolchains_discovery_requested_for_build_type = None
        
        for build_type in build_types:
            install_dir_abs_path = fs.path.join(working_dir_abs_path, self.get_install_dir_name(build_type))

            request_discovery = False
            if toolchains_discovery_requested_for_build_type is None:
                request_discovery = True
                toolchains_discovery_requested_for_build_type = build_type

            argv = [
                "conan", "install",
                "-s", f"build_type={self.base_to_conan_build_type[build_type]}",
                "-o", f"development={request_discovery}",
                f"--install-folder={install_dir_abs_path}",
                "--build=missing",
                f"{self.conanfile_path}"
            ]

            conan_install_invocations.append(argv)
            conan_install_build_types.append(build_type)

        if self.execute_conan:
            env = os.environ.copy()
            # env["CONAN_USER_HOME"] = fs.path.join(working_dir_abs_path, "conan_home")
            env["CONAN_ERROR_ON_OVERRIDE"] = "True"

            with importlib.resources.path(conan_generator, conan_export_resource) as conan_generator_path:
                self.logger.log_info("Invoking: {}".format(" ".join(conan_export_invocation)))
                self.process_runner.run(conan_export_invocation, env=env, check=True)

            for args in conan_install_invocations:
                self.logger.log_info("Invoking: {}".format(" ".join(args)))
                self.process_runner.run(args, env=env, check=True)

        with fs.open_fs(working_dir_abs_path) as work_fs:

            for build_type in conan_install_build_types:
                install_dir_name = self.get_install_dir_name(build_type)

                if build_type == toolchains_discovery_requested_for_build_type:
                    path = fs.path.join(install_dir_name, self.TOOLCHAIN_MANIFEST_FILENAME)
                    if not work_fs.exists(path):
                        raise RuntimeError(f"Expected generated file '{path}' does not exist.")

                path = fs.path.join(install_dir_name, self.THIRD_PARTY_MANIFEST_FILENAME)
                if not work_fs.exists(path):
                    raise RuntimeError(f"Expected generated file '{path}' does not exist.")

            for build_type in conan_install_build_types:
                install_dir_name = self.get_install_dir_name(build_type)

                if build_type == toolchains_discovery_requested_for_build_type:
                    toolchains_manifest_path = fs.path.join(install_dir_name, self.TOOLCHAIN_MANIFEST_FILENAME)
                    with work_fs.open(toolchains_manifest_path, "r") as f:
                        manifest = json.load(f)

                    for k, v in manifest.items():
                        if k not in self.toolchains_manifest.keys():
                            self.toolchains_manifest[k] = v

                # Gathe per-package 'imports' - dlls and dylibs required at runtime.
                package_dlls = {}

                for file in work_fs.glob(fs.path.join(install_dir_name, self.IMPORTS_BIN_DIRNAME, "*", "*.dll")):
                    abs_path = fs.path.join(working_dir_abs_path, fs.path.relpath(file.path))
                    package_name = fs.path.parts(file.path)[-2]
                    if package_name not in package_dlls:
                        package_dlls[package_name] = [abs_path]
                    else:
                        package_dlls[package_name].append(abs_path)

                for file in work_fs.glob(fs.path.join(install_dir_name, self.IMPORTS_LIB_DIRNAME, "*", "*.dylib")):
                    abs_path = fs.path.join(working_dir_abs_path, fs.path.relpath(file.path))
                    package_name = fs.path.parts(file.path)[-2]
                    if package_name not in package_dlls:
                        package_dlls[package_name] = [abs_path]
                    else:
                        package_dlls[package_name].append(abs_path)

                third_party_manifest_path = fs.path.join(install_dir_name, self.THIRD_PARTY_MANIFEST_FILENAME)
                with work_fs.open(third_party_manifest_path, "r") as f:
                    manifest = json.load(f)

                    for package_name in manifest.keys():
                        manifest[package_name]["load_time_libs"] = []

                    for package_name, paths in package_dlls.items():
                        manifest[package_name]["load_time_libs"] = paths

                    for k, v in manifest.items():
                        iface_key = base.build_target_key(k, base.TOOLCHAIN_DEFAULT, build_type)
                        assert (iface_key not in self.third_party_manifest.keys())
                        self.third_party_manifest[iface_key] = v

            with work_fs.open("debug_third_party_manifest.json", "w") as f:
                f.write(json.dumps(self.third_party_manifest, indent=4))

    def get_toolchains_manifest(self):
        return self.toolchains_manifest

    def get_third_party_manifest(self):

        third_party_manifest_2 = {}
        for k, v in self.third_party_manifest.items():
            third_party_manifest_2[k] = base.ThirdPartyTargetInterface.from_dict(v)

        return third_party_manifest_2



from __future__ import annotations

from conans import ConanFile, tools
from conans.model import Generator
from conans.errors import ConanInvalidConfiguration, ConanException
import traceback
import os
import json
import subprocess
import re


required_conan_version = ">=1.43 <=1.62"


def _sanitize_path(path: str):
    return os.path.normpath(path)


def _sanitize_paths(paths: list[str]):
    return [_sanitize_path(p) for p in paths]


class ToolchainDiscoveryError(RuntimeError):
    pass


class SemVer:
    def __init__(self, major: int, minor: int, patch: int):
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.patch}"


class ToolchainDiscoveryWindows:
    """
    This discovery procedure covers the most common setups and does not even try to be exhaustive in terms
    uniquely identifying all possible configurations:
    * Target arch is ignored in the toolchain name (key), since only x86_64 is supported. However, VS uses
      different binaries for targeting different platforms.
    * Clang on Windows can be installed by a lot of means, see: https://github.com/conan-io/conan/issues/10955
    * LLVM installation from llvm.org is not self-sufficient - MSVC needs to be installed to use it (it provides
      the Windows C++ language runtime and the MSVC standard library). Therefore, an example windows LLVM toolchain
      could potentially use different VS installations present.
      
    Warning!
    The resulting toolchain ids/names do not convey their full configuration/identity and thus generated binaries
    can be different on various machines. Toolchains discovered using this class should be used only for package
    development and never for building that package for deployment - this should be handled by tools curated
    by the Conan community and by fully respecting the Conan settings.

    """

    def __init__(self):
        self.toolchains = {}

    def dump_to_json(self):
        return json.dumps(self.toolchains, indent=4)

    # Settings allowed for users.
    # Whenever a new allowed setting is added, a detection mechanism must be added for it below.
    _USER_SUPPORTED_SETTINGS = {
        "compiler": {
            "Visual Studio": {
                "version": ["15", "16", "17"]
            },
            "clang": {
                "version": ["13"]
            }
        },
        "arch": ["x86_64"],
        "build_type": ["Debug", "Release"]
    }

    _OS_WINDOWS_KEY = "win"

    def _add_toolchain(self, build_os, name, arch, tc):
        key_includes_os = False
        key_includes_arch = False
        
        key = ""
        if key_includes_os:
            key += f"{build_os}_"
        key += f"{name}_"
        if key_includes_arch:
            key += f"{arch}_"
        key = key.removesuffix("_")
        
        self.toolchains[key] = tc

    def discover_toolchains(self, conanfile):
        # TODO: Get rid of conanfile argument - we discover everything we can, disregarding conanfile settings.
        # The argument is required for tools.vcvars_dict, even though we override the important options.
        # Replace tools.vcvars_dir with tools.vswhere.
        compilers = self._USER_SUPPORTED_SETTINGS["compiler"]

        vs_settings = compilers["Visual Studio"]
        valid_vc_vars = []

        for vs_version in vs_settings["version"]:
            # VS uses different toolchains for targeting different architectures.
            for arch in self._USER_SUPPORTED_SETTINGS["arch"]:
                try:
                    vc_vars = tools.vcvars_dict(conanfile, arch=arch, compiler_version=vs_version, filter_known_paths=True)
                except ConanException as e:
                    conanfile.output.info(f"Visual Studio version {vs_version} for arch {arch} -- not found.")
                else:
                    conanfile.output.info(f"Visual Studio version {vs_version} for arch {arch} -- found.")
                    valid_vc_vars.append((vs_version, arch, vc_vars))

        for vs_version, arch, vc_vars in valid_vc_vars:
            self._add_toolchain(self._OS_WINDOWS_KEY, f"vs{vs_version}_msvc", arch,
                                self._generate_toolchain_manifest_vs_msvc(vc_vars))

        clang_settings = compilers["clang"]

        clang_path = tools.which("clang")
        if clang_path is not None:

            try:
                args = [clang_path, "--version"]
                completed_proc = subprocess.run(args, check=True, capture_output=True, text=True)
                if completed_proc.stdout is not None:
                    output_lines = completed_proc.stdout.split(sep='\n')
                else:
                    raise ToolchainDiscoveryError("Cannot capture 'clang --version' output.")
            except subprocess.SubprocessError as e:
                raise ToolchainDiscoveryError("Could not execute 'clang --version'.")

            for line in output_lines:
                match = re.match(r"clang version (\d+)\.(\d+)\.(\d+)", line)
                if match is not None:
                    break
            else:
                raise ToolchainDiscoveryError("Could not parse 'clang --version' output.")

            major = match.group(1)
            minor = match.group(2)
            patch = match.group(3)
            conanfile.output.info(f"Compiler 'clang' -- found version {major}.{minor}.{patch}")

            supported_versions = clang_settings["version"]
            if major in supported_versions:
                # Go up twice
                llvm_root_path = os.path.split(os.path.split(clang_path)[0])[0]
                for vs_version, arch, vc_vars in valid_vc_vars:
                    self._add_toolchain(self._OS_WINDOWS_KEY, f"vs{vs_version}_llvm{major}", arch,
                                        self._generate_toolchain_manifest_vs_llvm(vc_vars, llvm_root_path,
                                                                                  SemVer(major, minor, patch)))
            else:
                vers = ", ".join(supported_versions)
                conanfile.output.info(f"Compiler 'clang' -- supported major versions are [{vers}], skipping")
        else:
            conanfile.output.info(f"Compiler 'clang' -- not found")

    @staticmethod
    def _generate_toolchain_manifest_vs_msvc(vc_vars):
        vc_tools_install_dir = _sanitize_path(vc_vars["VCToolsInstallDir"])
        vc_host_arch = vc_vars["VSCMD_ARG_HOST_ARCH"]
        vc_target_arch = vc_vars["VSCMD_ARG_TGT_ARCH"]

        vc_bin_path = os.path.join(f"{vc_tools_install_dir}", "bin", f"Host{vc_host_arch}", f"{vc_target_arch}")
        # vc_lib_path = os.path.join(f"{vc_tools_install_dir}", "lib", f"{vc_target_arch}")
        # vc_include_path = _sanitize_path(vc_tools_install_dir)

        toolchain = {
            "compiler_family": "msvc",
            "compiler_path": os.path.join(vc_bin_path, "cl.exe"),
            "librarian_path": os.path.join(vc_bin_path, "lib.exe"),
            "linker_path": os.path.join(vc_bin_path, "link.exe"),
        }

        # TODO: some of these have the version hardcoded
        msvc_extra_files = [
            'c1.dll',
            'c1xx.dll',
            'c2.dll',
            'atlprov.dll',  # Only needed if using ATL
            'msobj140.dll',
            'mspdb140.dll',
            'mspdbcore.dll',
            'mspdbsrv.exe',
            'mspft140.dll',
            'msvcp140.dll',
            'vcruntime140.dll',
            'tbbmalloc.dll',  # Required as of 16.2(14.22 .27905)
            os.path.join('1033', 'clui.dll'),
            os.path.join('1033', 'mspft140ui.dll'),  # Localized messages for static analysis
        ]

        msvc_extra_files = [os.path.join(vc_bin_path, p) for p in msvc_extra_files if len(p) != 0]

        toolchain["compiler_extra_files"] = _sanitize_paths(msvc_extra_files)
        toolchain["toolchain_include_dirs"] = _sanitize_paths(vc_vars["INCLUDE"])
        toolchain["toolchain_lib_dirs"] = _sanitize_paths(vc_vars["LIB"])

        return toolchain

    @staticmethod
    def _generate_toolchain_manifest_vs_llvm(vc_vars, llvm_root_path, llvm_semver):
        toolchain = {
            "compiler_family": "clang",
            "compiler_path": os.path.join(llvm_root_path, "bin", "clang++.exe"),
            "librarian_path": os.path.join(llvm_root_path, "bin", "llvm-ar.exe"),
            "linker_path": os.path.join(llvm_root_path, "bin", "lld-link.exe"),  # Must use lld-link on Windows.
            "compiler_extra_files": [],  # TODO: proobably some DLLs from the binaries dir should be added here.
        }

        # Clang on Windows links against MSVC's stdlib, runtime, etc.

        # The arguments and their order are based upon the --verbose output from the clang driver,
        # compiling a simple cpp file.

        # The clang include path must be first in the list.
        include_dirs = [os.path.join(llvm_root_path, "lib", "clang", str(llvm_semver), "include")]
        include_dirs.extend(_sanitize_paths(vc_vars["INCLUDE"]))

        lib_dirs = _sanitize_paths(vc_vars["LIB"])
        # The clang lib dir path must be last on the list.
        lib_dirs.append(os.path.join(llvm_root_path, "lib", "clang", str(llvm_semver), "lib", "windows"))

        toolchain["toolchain_include_dirs"] = include_dirs
        toolchain["toolchain_lib_dirs"] = lib_dirs

        return toolchain


def create_toolchain_discovery():
    detected_os = tools.detected_os()
    if detected_os == "Windows":
        return ToolchainDiscoveryWindows()
    else:
        raise NotImplementedError(f"Toolchain discovery not implemented on build OS: '{detected_os}'.")


# noinspection PyBroadException
class RapidGenerator(Generator):

    THIRD_PARTY_MANIFEST_FILENAME = "third_party_gen.json"
    TOOLCHAIN_MANIFEST_FILENAME = "toolchain_gen.json"

    # noinspection PyPropertyDefinition
    @property
    def filename(self):
        pass

    @property
    def content(self):
        try:

            if self.conanfile.options.development:
                self.conanfile.output.info("RapidGenerator: Running toolchain discovery...")

                td = create_toolchain_discovery()
                td.discover_toolchains(self.conanfile)

                return {
                    self.THIRD_PARTY_MANIFEST_FILENAME: self._generate_third_party_manifest_json(),
                    self.TOOLCHAIN_MANIFEST_FILENAME: td.dump_to_json()
                }
            else:
                return {
                    self.THIRD_PARTY_MANIFEST_FILENAME: self._generate_third_party_manifest_json(),
                }

        except Exception:
            traceback.print_exc()

    @staticmethod
    def _parse_definition(definition):
        # Preprocessor definition given on the command line.
        pattern = "^([_a-zA-Z][_a-zA-Z0-9]*)$"
        match = re.match(pattern, definition)
        if match is not None:
            name = match.group(1)
            return name, None

        # Preprocessor definition given on the command line, with value assignment.
        pattern = "^([_a-zA-Z][_a-zA-Z0-9]*)=(.*)$"
        match = re.match(pattern, definition)
        if match is not None:
            name = match.group(1)
            value = match.group(2)
            return name, value

    def _generate_third_party_manifest_json(self):
        deps_cpp_info = self.conanfile.deps_cpp_info
        targets_interfaces = {}
        for dep in deps_cpp_info.deps:
            dep_cpp_info = deps_cpp_info[dep]

            dep_interface = {
                "include_dirs": _sanitize_paths(dep_cpp_info.include_paths),
                "definitions": {},
                "link_libs": _sanitize_paths(dep_cpp_info.libs),
                "link_libs_dirs": _sanitize_paths(dep_cpp_info.lib_paths),
            }

            for d in dep_cpp_info.defines:
                name, value = self._parse_definition(d)
                dep_interface["definitions"][name] = value

            debug = False
            if debug:
                dep_interface["_impl"] = {
                    # cpp_info
                    "_includedirs": dep_cpp_info.includedirs,
                    "_libdirs": dep_cpp_info.libdirs,
                    "_resdirs": dep_cpp_info.resdirs,
                    "_bindirs": dep_cpp_info.bindirs,
                    "_builddirs": dep_cpp_info.builddirs,
                    "_libs": dep_cpp_info.libs,
                    "_defines": dep_cpp_info.defines,
                    "_cflags": dep_cpp_info.cflags,
                    "_cxxflags": dep_cpp_info.cxxflags,
                    "_sharedlinkflags": dep_cpp_info.sharedlinkflags,
                    "_exelinkflags": dep_cpp_info.exelinkflags,
                    "_frameworks": dep_cpp_info.frameworks,
                    "_frameworkdirs": dep_cpp_info.frameworkdirs,
                    "_system_libs": dep_cpp_info.system_libs,
                    # dep_cpp_info
                    "_include_paths": dep_cpp_info.include_paths,
                    "_lib_paths": dep_cpp_info.lib_paths,
                    "_bin_paths": dep_cpp_info.bin_paths,
                    "_build_paths": dep_cpp_info.build_paths,
                    "_res_paths": dep_cpp_info.res_paths,
                    "_framework_paths": dep_cpp_info.framework_paths,
                    "_build_modules_paths": dep_cpp_info.build_modules_paths,
                    "_version": dep_cpp_info.version,
                    # "_components": dep_cpp_info.components,
                }
                
            targets_interfaces[dep] = dep_interface
        return json.dumps(targets_interfaces, indent=4)


class Pkg(ConanFile):
    name = "rapidbuild"
    version = "0.5.2"
    license = "MIT"



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from enum import Enum, auto

import fs
import fs.errors

import bentoudev.dataclass.yaml_loader as yaml_loader

from . import base
from .compiler_base import (
    CompilerSettings,
    CompilerArgFormatter,
    _get_default_compiler_settings,
)
from .compiler_msvc import (
    CompilerArgFormatterMSVC,
    CompilerArgFormatterClangCL,
)
from .compiler_llvm import (
    CompilerArgFormatterGCC,
    CompilerArgFormatterClang
)
from .librarian_msvc import (
    LibrarianArgFormatterMSVC
)
from .librarian_llvm import (
    LibrarianArgFormatterGCC,
    LibrarianArgFormatterLLVM
)
from .linker_base import (
    LinkerSettings,
    _get_default_linker_settings
)
from .linker_msvc import (
    LinkerArgFormatterLink,
    LinkerArgFormatterLLDLink,
)
from .linker_llvm import (
    LinkerArgFormatterLDLLD,
    LinkerArgFormatterLD
)


class ToolchainSettingsWindowsDefault(base.ToolchainSettings):
    def get_build_types(self) -> set[str]:
        return {base.BUILD_TYPE_DEBUG, base.BUILD_TYPE_PROFILING, base.BUILD_TYPE_RELEASE}

    def get_compiler_settings(self, build_type: str) -> CompilerSettings:
        return _get_default_compiler_settings(build_type)

    def get_compiler_definitions(self, build_type: str) -> dict[str, base.Any]:
        return {}

    def get_linker_settings(self, build_type: str) -> LinkerSettings:
        return _get_default_linker_settings(build_type)


class ToolchainSettingsProviderHardcoded(base.ToolchainSettingsProvider):
    def __init__(self):
        self.d = {
            base.TOOLCHAIN_VS16_MSVC: ToolchainSettingsWindowsDefault(),
            base.TOOLCHAIN_VS17_MSVC: ToolchainSettingsWindowsDefault(),
            base.TOOLCHAIN_VS16_LLVM13: ToolchainSettingsWindowsDefault(),
            base.TOOLCHAIN_VS17_LLVM13: ToolchainSettingsWindowsDefault(),
        }

    def get_toolchain_settings(self, toolchain_name) -> base.ToolchainSettings:
        try:
            return self.d[toolchain_name]
        except KeyError:
            raise NotImplementedError()

    def get_toolchains(self) -> set[str]:
        return set(self.d.keys())


@dataclass
class ToolchainBuildTypeSettings:
    compiler_settings: CompilerSettings = field(default_factory=CompilerSettings)
    compiler_defs: dict[str, base.Any] = field(default_factory=dict)
    linker_settings: LinkerSettings = field(default_factory=LinkerSettings)


@dataclass
class ToolchainSettingsData(base.ToolchainSettings):
    build_types: dict[str, ToolchainBuildTypeSettings] = field(default_factory=dict)

    def _get_build_type(self, build_type: str) -> ToolchainBuildTypeSettings:
        try:
            return self.build_types[build_type]
        except KeyError:
            raise NotImplementedError()

    def get_build_types(self) -> set[str]:
        return set(self.build_types.keys())

    def get_compiler_settings(self, build_type: str) -> CompilerSettings:
        return self._get_build_type(build_type).compiler_settings

    def get_compiler_definitions(self, build_type: str) -> dict[str, base.Any]:
        return self._get_build_type(build_type).compiler_defs

    def get_linker_settings(self, build_type: str) -> LinkerSettings:
        return self._get_build_type(build_type).linker_settings


class ToolchainSettingsDeserializer(ABC):
    @abstractmethod
    def deserialize(self) -> ToolchainSettingsData | None:
        pass


class ToolchainSettingsSettingsDeserializerFromFile(ToolchainSettingsDeserializer):
    def __init__(self, *, wks_fs, path, logger):
        self.fs = wks_fs
        self.file_path = path
        self.logger = logger

    def deserialize(self) -> ToolchainSettingsData | None:
        file_contents = ""
        try:
            with self.fs.open(self.file_path, 'r') as f:
                file_contents = f.read()
        except fs.errors.ResourceNotFound:
            self.logger.log_error(f"File '{self.file_path}' does not exist.")
            return None

        try:
            settings = yaml_loader.load_yaml_dataclass(
                ToolchainSettingsData, fs.path.basename(self.file_path), file_contents)
            if settings is None:
                self.logger.log_error(f"File '{self.file_path}' is not a valid yaml file.")
            return settings
        except yaml_loader.DataclassLoadError as err:
            self.logger.log_error(f"File '{self.file_path}':\n{str(err)}.")
            return None


class ToolchainSettingsDeseralizeError(RuntimeError):
    pass


class ToolchainSettingsProviderFromFile(base.ToolchainSettingsProvider):
    def __init__(self, wks_fs, path, logger):
        self.deserializer = ToolchainSettingsSettingsDeserializerFromFile(wks_fs=wks_fs, path=path, logger=logger)

        self.settings = self.deserializer.deserialize()
        if self.settings is None:
            raise ToolchainSettingsDeseralizeError()

        self.d = {
            base.TOOLCHAIN_VS16_MSVC: self.settings,
            base.TOOLCHAIN_VS17_MSVC: self.settings,
            base.TOOLCHAIN_VS16_LLVM13: self.settings,
            base.TOOLCHAIN_VS17_LLVM13: self.settings,
        }

    def get_toolchain_settings(self, toolchain_name) -> base.ToolchainSettings:
        try:
            return self.d[toolchain_name]
        except KeyError:
            raise NotImplementedError()

    def get_toolchains(self) -> set[str]:
        return set(self.d.keys())


class ToolchainSettingsProviderDefault(base.ToolchainSettingsProvider):
    TOOLCHAIN_SETTINGS_FILE = "toolchain_settings.yml"

    def __init__(self, wks_fs, logger):
        try:
            path = fs.path.join("src", self.TOOLCHAIN_SETTINGS_FILE)
            self.provider = ToolchainSettingsProviderFromFile(wks_fs, path, logger)
            logger.log_info(f"Using toolchain settings from '{path}'.")
        except ToolchainSettingsDeseralizeError:
            self.provider = ToolchainSettingsProviderHardcoded()
            logger.log_info("Using default hardcoded toolchain settings.")

    def get_toolchain_settings(self, toolchain_name) -> base.ToolchainSettings:
        return self.provider.get_toolchain_settings(toolchain_name)

    def get_toolchains(self) -> set[str]:
        return self.provider.get_toolchains()


# noinspection PyArgumentList
class CompilerFamily(Enum):
    MSVC = auto()
    CLANG = auto()
    CLANG_CL = auto()
    GCC = auto()


class CouldNotDetermineCompilerFamily(RuntimeError):
    pass


def guess_compiler_family_from_compiler_path(path: str) -> CompilerFamily:
    filename = Path(path).parts[-1]
    if filename == "cl" or filename == "cl.exe":
        return CompilerFamily.MSVC
    if filename == "clang" or filename == "clang.exe" or filename == "clang++" or filename == "clang++.exe":
        return CompilerFamily.CLANG
    if filename == "clang-cl" or filename == "clang-cl.exe":
        return CompilerFamily.CLANG_CL
    if filename == "gcc":
        return CompilerFamily.GCC
    raise CouldNotDetermineCompilerFamily(f"compiler path: {path}")


def create_compiler_arg_formatter(f: CompilerFamily) -> CompilerArgFormatter:
    return {
        CompilerFamily.MSVC: CompilerArgFormatterMSVC(),
        CompilerFamily.CLANG: CompilerArgFormatterClang(),
        CompilerFamily.CLANG_CL: CompilerArgFormatterClangCL(),
        CompilerFamily.GCC: CompilerArgFormatterGCC()
    }[f]


def get_compiler_system_definitions(f: CompilerFamily, build_type: str):
    defs = {
        base.BUILD_TYPE_MATCH_ALL: {
            "_MT": None,
            "_DLL": None,
        },
        base.BUILD_TYPE_DEBUG: {
            "_DEBUG": None,
        },
        base.BUILD_TYPE_RELEASE: {
        },
        base.BUILD_TYPE_PROFILING: {
        },
    }
    try:
        return defs[base.BUILD_TYPE_MATCH_ALL] | defs[build_type]
    except KeyError:
        raise NotImplementedError()


# noinspection PyArgumentList
class LibrarianFamily(Enum):
    MSVC = auto()
    LLVM = auto()
    GCC = auto()


class CouldNotDetermineLibrarianFamily(RuntimeError):
    pass


def guess_librarian_family_from_path(path: str) -> LibrarianFamily:
    filename = Path(path).parts[-1]
    if filename == "lib" or filename == "lib.exe":
        return LibrarianFamily.MSVC
    if filename == "llvm-ar" or filename == "llvm-ar.exe":
        return LibrarianFamily.LLVM
    if filename == "ar" or filename == "ar.exe":
        return LibrarianFamily.GCC
    raise CouldNotDetermineLibrarianFamily(f"librarian path: {path}")


def create_librarian_arg_formatter(f: LibrarianFamily):
    return {
        LibrarianFamily.MSVC: LibrarianArgFormatterMSVC(),
        LibrarianFamily.LLVM: LibrarianArgFormatterLLVM(),
        LibrarianFamily.GCC: LibrarianArgFormatterGCC(),
    }[f]


# noinspection PyArgumentList
class LinkerFamily(Enum):
    LINK = auto()
    LLD_LINK = auto()
    LD_LLD = auto()
    LD = auto()


class CouldNotDetermineLinkerFamily(RuntimeError):
    pass


def guess_linker_family_from_path(path: str) -> LinkerFamily:
    filename = Path(path).parts[-1]
    if filename == "link" or filename == "link.exe":
        return LinkerFamily.LINK
    if filename == "lld-link" or filename == "lld-link.exe":
        return LinkerFamily.LLD_LINK
    if filename == "ld.lld" or filename == "ld.lld.exe":
        return LinkerFamily.LD_LLD
    if filename == "ld" or filename == "ld.exe":
        return LinkerFamily.LD
    raise CouldNotDetermineLinkerFamily(f"linker path: {path}")


def create_linker_arg_formatter(f: LinkerFamily):
    return {
        LinkerFamily.LINK: LinkerArgFormatterLink(),
        LinkerFamily.LLD_LINK: LinkerArgFormatterLLDLink(),
        LinkerFamily.LD_LLD: LinkerArgFormatterLDLLD(),
        LinkerFamily.LD: LinkerArgFormatterLD()
    }[f]


def get_linker_system_libs(f: LinkerFamily, build_type: str):
    debug_libs = [
        # THE ORDER OF THESE IMPORT LIBS IS IMPORTANT
        # Switching msvcrtd.lib with vcruntimed.lib results in a duplicate symbol error with lld-link.

        # DLL import library for the Debug version of the UCRT (ucrtbased.dll).
        'ucrtd.lib',
        # Static library for the Debug version of the native CRT startup for use with DLL UCRT and vcruntime.
        'msvcrtd.lib',
        # DLL import library for the Debug vcruntime (vcruntime<version>d.dll).
        'vcruntimed.lib',
        # Multithreaded, dynamic link (import library for msvcp<version>d.dll)
        'msvcprtd.lib',
    ]
    release_libs = [
        # DLL import library for the UCRT (ucrtdll).
        'ucrt.lib',
        # Static library for the native CRT startup for use with DLL UCRT and vcruntime.
        'msvcrt.lib',
        # DLL import library for the vcruntime (vcruntime<version>.dll).
        'vcruntime.lib',
        # Multithreaded, dynamic link (import library for msvcp<version>.dll)
        'msvcprt.lib',
    ]
    libs = {
        base.BUILD_TYPE_MATCH_ALL: [
            'kernel32.lib',
            'user32.lib',
            'gdi32.lib',
            'shell32.lib',
            'winspool.lib',
            'ole32.lib',
            'oleaut32.lib',
            'uuid.lib',
            'comdlg32.lib',
            'advapi32.lib',
        ],
        base.BUILD_TYPE_DEBUG: debug_libs,
        base.BUILD_TYPE_RELEASE: release_libs,
        base.BUILD_TYPE_PROFILING: release_libs,
    }
    try:
        return libs[base.BUILD_TYPE_MATCH_ALL] + libs[build_type]
    except KeyError:
        raise NotImplementedError()








class GraphCreationFailure(RuntimeError):
    pass


class GraphCreationVerticesNonUnique(GraphCreationFailure):
    pass


class GraphCreationMissingVertex(GraphCreationFailure):
    pass


class GraphAccessInvalid(RuntimeError):
    pass


class GraphInvalidVertex(GraphAccessInvalid):
    pass


class DirectedGraph:
    def __init__(self, *, vertices: list[int], edges: list[tuple[int, int]]):
        if len(set(vertices)) != len(vertices):
            raise GraphCreationVerticesNonUnique("Vertex list elements must be unique.")

        self._adj_lists = {v: [] for v in vertices}
        for src, dst in edges:
            if src not in vertices:
                raise GraphCreationMissingVertex(f"Vertex '{src}' present in edge list does not exist.")
            if dst not in vertices:
                raise GraphCreationMissingVertex(f"Vertex '{dst}' present in edge list does not exist.")
            self._adj_lists[src].append(dst)

    def vertices(self) -> list[int]:
        return list(self._adj_lists.keys())

    def adjacent(self, vert: int) -> list[int]:
        if vert not in self.vertices():
            raise GraphInvalidVertex(f"Vertex '{vert}' does not exist.")
        return self._adj_lists[vert]

    def _is_cyclic_impl(self, v, visited, rec_stack):

        # Mark current node as visited and
        # adds to recursion stack
        visited[v] = True
        rec_stack[v] = True

        # Recur for all neighbours
        # if any neighbour is visited and in
        # rec_stack then graph is cyclic
        for neighbour in self.adjacent(v):
            if not visited[neighbour]:
                if self._is_cyclic_impl(neighbour, visited, rec_stack):
                    return True
            elif rec_stack[neighbour]:
                return True

        # The node needs to be popped from
        # recursion stack before function ends
        rec_stack[v] = False
        return False

    def is_cyclic(self):
        vert_count = len(self.vertices())
        visited = [False] * (vert_count + 1)
        rec_stack = [False] * (vert_count + 1)
        for vert in range(vert_count):
            if not visited[vert]:
                if self._is_cyclic_impl(vert, visited, rec_stack):
                    return True
        return False





from dataclasses import dataclass, field
from typing import Any
import fs

from .workspace_def import (
    WorkspaceDefinition, TargetKind
)
from .graph import (
    DirectedGraph
)
from . import base
from .base import (
    extend_unique,
    append_unique
)


class TargetNotFoundError(RuntimeError):
    pass


class TargetDependencyError(RuntimeError):
    pass


@dataclass
class TargetInterface:
    name: str
    kind: TargetKind
    include_dirs: list[str] = field(default_factory=list)
    definitions: dict[str, Any] = field(default_factory=dict)
    link_libs: list[str] = field(default_factory=list)
    link_libs_external: list[str] = field(default_factory=list)
    link_libs_external_dirs: list[str] = field(default_factory=list)
    load_time_libs: list[str] = field(default_factory=list)
    load_time_libs_external: list[str] = field(default_factory=list)


@dataclass
class TargetImplementation:
    name: str
    kind: TargetKind
    source_dir: str = ""
    include_dirs: list[str] = field(default_factory=list)
    definitions: dict[str, Any] = field(default_factory=dict)
    link_libs: list[str] = field(default_factory=list)
    link_libs_external: list[str] = field(default_factory=list)
    link_libs_external_dirs: list[str] = field(default_factory=list)
    load_time_libs: list[str] = field(default_factory=list)
    load_time_libs_external: list[str] = field(default_factory=list)


class TargetGraph:
    def __init__(self, wks_def: WorkspaceDefinition = WorkspaceDefinition()):
        self.wks_def = wks_def
        self.graph = self._build()

        self.targets_interfaces = {}
        self.targets_impls = {}

    def _build(self):

        targets = self.wks_def.targets

        targets_ids = {tgt.name: idx for idx, tgt in enumerate(targets)}

        edges = []
        for target in targets:
            for dep_name in target.private_props.dependencies:
                if dep_name not in targets_ids.keys():
                    raise TargetNotFoundError(f"Target '{dep_name}' privately referenced by '{target.name}' was not defined.")
                edges.append((targets_ids[target.name], targets_ids[dep_name]))

            for dep_name in target.public_props.dependencies:
                if dep_name not in targets_ids.keys():
                    raise TargetNotFoundError(f"Target '{dep_name}' publicly referenced by '{target.name}' was not defined.")
                edges.append((targets_ids[target.name], targets_ids[dep_name]))

        return DirectedGraph(vertices=list(targets_ids.values()), edges=edges)

    def is_valid(self):
        return not self.graph.is_cyclic()

    @staticmethod
    def _add_definitions(iface, impl, dep_defs, *, is_public_dep):
        for name, value in dep_defs.items():
            # TODO: resolve conflicts (A=3, A=4)
            if is_public_dep:
                if name not in iface.definitions:
                    iface.definitions[name] = value
            if name not in impl.definitions:
                impl.definitions[name] = value

    @staticmethod
    def _process_dependency(iface, impl, dep_iface, dep_simple_name, *, is_public_dep):
        if dep_iface.kind == TargetKind.EXECUTABLE:
            raise TargetDependencyError("Cannot declare dep on an exe.")

        TargetGraph._add_definitions(iface, impl, dep_iface.definitions, is_public_dep=is_public_dep)

        if is_public_dep:
            extend_unique(iface.include_dirs, dep_iface.include_dirs)
        extend_unique(impl.include_dirs, dep_iface.include_dirs)

        extend_unique(iface.link_libs, dep_iface.link_libs)
        extend_unique(impl.link_libs, dep_iface.link_libs)

        extend_unique(iface.link_libs_external, dep_iface.link_libs_external)
        extend_unique(impl.link_libs_external, dep_iface.link_libs_external)

        extend_unique(iface.link_libs_external_dirs, dep_iface.link_libs_external_dirs)
        extend_unique(impl.link_libs_external_dirs, dep_iface.link_libs_external_dirs)

        if dep_iface.kind == TargetKind.DYNAMIC_LIB:
            append_unique(iface.load_time_libs, dep_simple_name)
            append_unique(impl.load_time_libs, dep_simple_name)

        extend_unique(iface.load_time_libs, dep_iface.load_time_libs)
        extend_unique(impl.load_time_libs, dep_iface.load_time_libs)

        extend_unique(iface.load_time_libs_external, dep_iface.load_time_libs_external)
        extend_unique(impl.load_time_libs_external, dep_iface.load_time_libs_external)

    @staticmethod
    def _process_ext_dependency(iface, impl, dep_iface, *, is_public_dep):

        TargetGraph._add_definitions(iface, impl, dep_iface.definitions, is_public_dep=is_public_dep)

        if is_public_dep:
            extend_unique(iface.include_dirs, dep_iface.include_dirs)
        extend_unique(impl.include_dirs, dep_iface.include_dirs)

        extend_unique(iface.link_libs_external, dep_iface.link_libs)
        extend_unique(impl.link_libs_external, dep_iface.link_libs)

        extend_unique(iface.link_libs_external_dirs, dep_iface.link_libs_dirs)
        extend_unique(impl.link_libs_external_dirs, dep_iface.link_libs_dirs)

        extend_unique(iface.load_time_libs_external, dep_iface.load_time_libs)
        extend_unique(impl.load_time_libs_external, dep_iface.load_time_libs)

    def calc_target_iface_impl(self, target_name: str, source_dir_abs_path: str, toolchain_name: str, build_type: str,
                               third_party_manifest):
        assert self.is_valid()

        target_def = self.wks_def.find_target_def(target_name)
        if target_def is None:
            raise TargetNotFoundError()

        if target_def.name in self.targets_impls:
            return self.targets_impls[target_def.name], self.targets_interfaces.get(target_def.name, None)

        target_key = base.build_target_key(target_def.name, toolchain_name, build_type)

        iface = TargetInterface(name=target_key, kind=target_def.kind)
        impl = TargetImplementation(name=target_key, kind=target_def.kind, source_dir="")

        target_public_include_dir_name = "include"
        target_private_include_dir_name = "src"
        target_source_dir_name = "src"

        target_public_include_path = fs.path.join(source_dir_abs_path, target_def.name, target_public_include_dir_name)
        target_private_include_path = fs.path.join(source_dir_abs_path, target_def.name, target_private_include_dir_name)
        target_source_dir_path = fs.path.join(source_dir_abs_path, target_def.name, target_source_dir_name)

        iface.include_dirs.append(target_public_include_path)
        impl.include_dirs.append(target_public_include_path)
        impl.include_dirs.append(target_private_include_path)

        impl.source_dir = target_source_dir_path

        iface.link_libs.append(target_def.name)

        #
        # process preprocessor definitions
        #

        # Own public definitions -> interface definitions + impl definitions
        self._add_definitions(iface, impl, target_def.public_props.definitions, is_public_dep=True)
        self._add_definitions(iface, impl, target_def.private_props.definitions, is_public_dep=False)

        #
        # process dependencies
        #

        for dep_name in target_def.public_props.dependencies:
            dep_iface, _ = self.calc_target_iface_impl(dep_name, source_dir_abs_path, toolchain_name, build_type, third_party_manifest)

            self._process_dependency(iface, impl, dep_iface, dep_name, is_public_dep=True)

        for dep_name in target_def.private_props.dependencies:
            dep_iface, _ = self.calc_target_iface_impl(dep_name, source_dir_abs_path, toolchain_name, build_type, third_party_manifest)

            self._process_dependency(iface, impl, dep_iface, dep_name, is_public_dep=False)

        #
        # process external dependencies
        #

        for ext_name in target_def.public_props.externals:
            dep_iface = third_party_manifest[base.build_target_key(ext_name, base.TOOLCHAIN_DEFAULT, build_type)]

            self._process_ext_dependency(iface, impl, dep_iface, is_public_dep=True)

        for ext_name in target_def.private_props.externals:
            dep_iface = third_party_manifest[base.build_target_key(ext_name, base.TOOLCHAIN_DEFAULT, build_type)]

            self._process_ext_dependency(iface, impl, dep_iface, is_public_dep=False)

        # save result
        target_key = base.build_target_key(target_def.name, toolchain_name, build_type)

        # Executables don't need an interface
        if target_def.kind != TargetKind.EXECUTABLE:
            self.targets_interfaces[target_key] = iface
        self.targets_impls[target_key] = impl

        return iface, impl


def create_targets_interfaces_and_implementations(
        source_dir_abs_path: str, toolchains: list[str], build_types: list[str], third_party_manifest,
        workspace_def: WorkspaceDefinition
):
    assert (len(source_dir_abs_path) != 0)
    assert (len(build_types) != 0)
    assert (len(toolchains) != 0)

    target_graph = TargetGraph(wks_def=workspace_def)
    if not target_graph.is_valid():
        raise TargetDependencyError()

    for toolchain_name in toolchains:
        for build_type in build_types:
            for target_def in workspace_def.targets:
                target_graph.calc_target_iface_impl(target_def.name, source_dir_abs_path, toolchain_name,
                                                    build_type, third_party_manifest)

    return target_graph.targets_interfaces, target_graph.targets_impls



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto

from . import base
from .workspace_def import TargetKind


class LinkerWarningState(Enum):
    ENABLE = auto()
    DISABLE = auto()
    TREAT_AS_ERROR = auto()


@dataclass
class LinkerWarningSettings:
    treat_as_errors: bool = True
    warning_states: dict[str, LinkerWarningState] = field(default_factory=dict)


@dataclass
class LinkerSettings:
    warning_settings: LinkerWarningSettings = field(default_factory=LinkerWarningSettings)
    enable_debug_info: bool = False
    enable_security_checks: bool = False
    enable_optimizations: bool = False


class LinkerArgFormatter(ABC):
    @abstractmethod
    def format_input_output_declaration(
            self, output_dir_path: str, output_base_name: str, target_kind: TargetKind
    ) -> list[str]:
        pass

    @abstractmethod
    def format_lib_arg(self, lib: str) -> str:
        pass

    @abstractmethod
    def format_lib_dir_arg(self, lib_dir: str) -> str:
        pass

    @abstractmethod
    def format_settings_args(self, settings: LinkerSettings) -> list[str]:
        pass


def _get_default_linker_settings(build_type: str) -> LinkerSettings:
    settings = LinkerSettings()
    settings.warning_settings.treat_as_errors = True
    settings.warning_settings.warning_states["4099"] = LinkerWarningState.DISABLE

    if build_type == base.BUILD_TYPE_DEBUG:
        settings.enable_debug_info = True
        settings.enable_security_checks = True
        settings.enable_optimizations = False
        return settings

    if build_type == base.BUILD_TYPE_RELEASE:
        settings.enable_debug_info = False
        settings.enable_security_checks = False
        settings.enable_optimizations = True
        return settings

    if build_type == base.BUILD_TYPE_PROFILING:
        settings.enable_debug_info = True
        settings.enable_security_checks = False
        settings.enable_optimizations = True
        return settings

    raise NotImplementedError()





#!/usr/bin/env
from typing import Any, Union

import bentoudev.dataclass.yaml_loader
import yaml
import os.path
import fs.path
from fs.base import FS
from fs.osfs import OSFS
import fs.errors

from . import base
from .conan import SystemConfigManifestProviderConan
from .fastbuild import BuildScriptEmitterFastbuild
from .target_graph import create_targets_interfaces_and_implementations
from .toolchain import ToolchainSettingsProviderDefault
from .workspace_def import (
    WorkspaceDefinition,
    WorkspaceInvalidFormat,
)


def _validate_third_party_manifest(manifest):
    if not isinstance(manifest, dict):
        return False
    return True


def _validate_toolchains_manifest(manifest, supported_toolchains):
    if not isinstance(manifest, dict):
        return False
    if len(manifest.keys()) == 0:
        return False
    return True


def _validate_toolchains_settings(settings):
    if not isinstance(settings, dict):
        return False
    if len(settings.keys()) == 0:
        return False
    return True


class Workspace:
    def __init__(self, *,
                 name: str,
                 wks_fs: FS,
                 logger: base.Logger,
                 system_config_provider: base.SystemConfigManifestProvider,
                 toolchain_settings_provider: base.ToolchainSettingsProvider,
                 build_script_emitter: base.BuildScriptEmitter
                 ):

        # TODO: sanitize name string: only alphanum + _, no spaces
        self.name = name
        self.fs = wks_fs

        self.configure_dir_name = "configure"
        self.source_dir_name = "src"
        self.build_dir_name = "build"

        self.workspace_def: Union[WorkspaceDefinition, None] = None
        self.third_party_manifest: dict[str, Any] = {}
        self.toolchains_manifest: dict[str, Any] = {}

        self.build_types = [
            base.BUILD_TYPE_DEBUG,
            base.BUILD_TYPE_RELEASE,
            base.BUILD_TYPE_PROFILING
        ]
        self.supported_toolchains = [
            base.TOOLCHAIN_VS16_MSVC,
            base.TOOLCHAIN_VS16_LLVM13,
            base.TOOLCHAIN_VS17_MSVC,
            base.TOOLCHAIN_VS17_LLVM13,
        ]

        self.logger = logger
        self.system_config_provider = system_config_provider
        self.toolchain_settings_provider = toolchain_settings_provider
        self.build_script_emitter = build_script_emitter

    def __del__(self):
        self.fs.close()
        pass

    def _get_source_dir_path(self):
        return self.source_dir_name

    def _get_source_dir_abs_path(self):
        if self.fs.hassyspath(self.source_dir_name):
            return self.fs.getsyspath(self.source_dir_name)
        return f"wks://{self.source_dir_name}"

    def _get_configure_dir_abs_path(self):
        if self.fs.hassyspath(self.configure_dir_name):
            return self.fs.getsyspath(self.configure_dir_name)
        return f"wks://{self.configure_dir_name}"

    def _get_build_dir_abs_path(self):
        if self.fs.hassyspath(self.build_dir_name):
            return self.fs.getsyspath(self.build_dir_name)
        return f"wks://{self.build_dir_name}"

    TARGETS_FILENAME = "rapid_targets.yml"
    USE_BENTOUDEV_LOADER = False # TODO: Current yaml structure is too far off from the code data structures

    def configure(self) -> bool:

        #
        # Load workspace target definitions
        #

        targets_file_path = fs.path.join(self.source_dir_name, self.TARGETS_FILENAME)
        self.logger.log_info(f"Reading workspace definition '{targets_file_path}'...")


        if self.USE_BENTOUDEV_LOADER:
            file_contents = ""
            try:
                with self.fs.open(targets_file_path, 'r') as f:
                    file_contents = f.read()
            except fs.errors.ResourceNotFound:
                self.logger.log_error(f"File '{targets_file_path}' does not exist.")
                return False

            try:
                try:
                    self.workspace_def = bentoudev.dataclass.yaml_loader.load_yaml_dataclass(
                        WorkspaceDefinition, self.TARGETS_FILENAME, file_contents)
                except bentoudev.dataclass.yaml_loader.DataclassLoadError as err:
                    self.logger.log_error(f"Error location line '{err.source.line_number}', column '{err.source.column_number}'.")
                    raise WorkspaceInvalidFormat()
            except WorkspaceInvalidFormat:
                self.logger.log_error(f"File '{targets_file_path}' is not a valid targets definition file.")
                return False

        else:
            try:
                with self.fs.open(targets_file_path, 'r') as f:
                    if not yaml.__with_libyaml__:
                        self.logger.log_info("Warning: using yaml module without C bindings.")
                    targets_defs = yaml.load(f, yaml.SafeLoader)
            except fs.errors.ResourceNotFound:
                self.logger.log_error(f"File '{targets_file_path}' does not exist.")
                return False

            try:
                self.workspace_def = WorkspaceDefinition.from_dict(targets_defs)
            except WorkspaceInvalidFormat:
                self.logger.log_error(f"File '{targets_file_path}' is not a valid targets definition file.")
                return False

        self.logger.log_info("Workspace definition loaded.")

        # print(json.dumps(target_defs, indent=2))

        #
        # Prepare configuration directory, provide manifests
        #
        self.fs.makedir(self.configure_dir_name, recreate=True)

        self.system_config_provider.run(self._get_configure_dir_abs_path(), self.build_types)

        self.third_party_manifest = self.system_config_provider.get_third_party_manifest()

        if not _validate_third_party_manifest(self.third_party_manifest):
            self.logger.log_error("Provided third party manifest is invalid.")
            return False

        self.toolchains_manifest = self.system_config_provider.get_toolchains_manifest()

        if not _validate_toolchains_manifest(self.toolchains_manifest, self.supported_toolchains):
            self.logger.log_error("Provided toolchains manifest is invalid.")
            return False

        manifest_toolchains = self.toolchains_manifest.keys()
        if not set(manifest_toolchains).issubset(self.supported_toolchains):
            self.logger.log_error("Manifest toolchains must be a subset of supported toolchains.")
            self.logger.log_error(f"Toolchains in manifest: {sorted(manifest_toolchains)}")
            self.logger.log_error(f"Supported toolchains: {sorted(self.supported_toolchains)}")
            return False

        settings_toolchains = self.toolchain_settings_provider.get_toolchains()
        if not set(manifest_toolchains).issubset(settings_toolchains):
            self.logger.log_error("Settings must be provided for all toolchains from the toolchain manifest.")
            self.logger.log_error(f"Manifest toolchains: {sorted(list(manifest_toolchains))}")
            self.logger.log_error(f"Toolchains for which settings were provided : {sorted(list(settings_toolchains))}")
            return False

        ok = True
        for toolchain in self.toolchain_settings_provider.get_toolchains():
            settings = self.toolchain_settings_provider.get_toolchain_settings(toolchain)
            if settings.get_build_types() != set(self.build_types):
                self.logger.log_error("Settings for all supported build types must be provided.")
                self.logger.log_error(f"Build types in settings for toolchain {toolchain}: "
                                      f"{sorted(list(settings.get_build_types()))}")
                self.logger.log_error(f"Supported build types: {sorted(self.build_types)}")
                ok = False
        if not ok:
            return False


        #
        # Build targets implementations
        #
        targets_interfaces, targets_impls = create_targets_interfaces_and_implementations(
            self._get_source_dir_abs_path(),
            self.toolchains_manifest.keys(),
            self.build_types,
            self.third_party_manifest,
            self.workspace_def
        )

        assert (isinstance(targets_interfaces, dict))
        assert (isinstance(targets_impls, dict) and len(targets_impls.keys()) != 0)

        #
        # Generate build script for targets based on implementations
        #

        build_script_filename = self.build_script_emitter.filename()
        build_script_contents = self.build_script_emitter.contents(
            self.name,
            self._get_source_dir_abs_path(),
            self._get_configure_dir_abs_path(),
            self._get_build_dir_abs_path(),
            self.toolchains_manifest,
            self.toolchain_settings_provider,
            self.build_types,
            [d.name for d in self.workspace_def.targets],
            targets_impls
        )

        build_script_path = fs.path.join(self.configure_dir_name, build_script_filename)

        self.logger.log_info(f"Writing build script '{build_script_path}'...")
        with self.fs.open(build_script_path, 'w') as f:
            f.write(build_script_contents)

        self.logger.log_info("Configuring done.")
        return True


def open_workspace(wks_dir: str, *, name=None, dev_mode=False) -> Workspace:
    logger = base.LoggerDefault()

    wks_fs = OSFS(os.path.abspath(wks_dir))
    wks_name = name if name is not None else wks_dir

    logger.log_info(f"Opening workspace '{name if name is not None else wks_dir}' (dir: '{wks_dir}')...")

    system_config_provider = SystemConfigManifestProviderConan(
        logger=logger,
        process_runner=base.ProcessRunnerDefault(),
        conanfile_path=wks_fs.getsyspath("conanfile.py"),
        execute_conan=not dev_mode,
        base_to_conan_build_type={
            base.BUILD_TYPE_DEBUG: "Debug",
            base.BUILD_TYPE_RELEASE: "Release",
            base.BUILD_TYPE_PROFILING: "Release",
        }
    )

    toolchain_settings_provider = ToolchainSettingsProviderDefault(wks_fs, logger)

    wks = Workspace(
        name=wks_name,
        wks_fs=wks_fs,
        logger=logger,
        system_config_provider=system_config_provider,
        toolchain_settings_provider=toolchain_settings_provider,
        build_script_emitter=BuildScriptEmitterFastbuild()
    )
    return wks



from .librarian_base import LibrarianArgFormatter


class LibrarianArgFormatterMSVC(LibrarianArgFormatter):
    def format_input_output_args(self) -> list[str]:
        return ["/NOLOGO", "/MACHINE:X64", "/WX", "\"%1\"", "/OUT:\"%2\""]




from enum import Enum, auto
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Any

from . import base


class LanguageStandard(Enum):
    C89 = auto()
    C99 = auto()
    C11 = auto()
    C17 = auto()
    C23 = auto()
    C_LATEST = auto()
    CPP03 = auto()
    CPP11 = auto()
    CPP14 = auto()
    CPP17 = auto()
    CPP20 = auto()
    CPP23 = auto()
    CPP_LATEST = auto()


class UnsupportedSettingError(NotImplementedError):
    pass


class UnsupportedLanguageError(UnsupportedSettingError):
    pass


class LanguageConformance(Enum):
    MIN = auto()
    MAX = auto()


class CompilerWarningState(Enum):
    ENABLE = auto()
    DISABLE = auto()
    REPORT_ONCE = auto()
    TREAT_AS_ERROR = auto()


class CompilerWarningLevel(Enum):
    W0 = auto()
    W1 = auto()
    W2 = auto()
    W3 = auto()
    W4 = auto()
    All = auto()


@dataclass
class CompilerWarningSettings:
    level: CompilerWarningLevel = CompilerWarningLevel.W3
    treat_as_errors: bool = True
    version: str = ""
    warning_states: dict[str, CompilerWarningState] = field(default_factory=dict)


class CompilerFloatBehavior(Enum):
    STRICT = auto()
    PRECISE = auto()
    FAST = auto()


class CompilerOptimizationMode(Enum):
    NONE = auto()
    MINIMIZE_SIZE = auto()
    MAXIMIZE_SPEED = auto()


@dataclass
class CompilerSettings:
    lang_std: LanguageStandard = LanguageStandard.CPP17
    lang_conformance: LanguageConformance = LanguageConformance.MAX
    warning_settings: CompilerWarningSettings = field(default_factory=CompilerWarningSettings)
    enable_rtti: bool = False
    enable_exceptions: bool = False
    float_behavior: CompilerFloatBehavior = CompilerFloatBehavior.PRECISE
    use_debug_runtime: bool = False
    enable_debug_info: bool = False
    optimization_mode: CompilerOptimizationMode = CompilerOptimizationMode.NONE
    enable_security_checks: bool = False


class CompilerArgFormatter(ABC):
    @abstractmethod
    def format_beginning_args(self) -> list[str]:
        pass

    @abstractmethod
    def format_object_compilation_args(self) -> list[str]:
        pass

    @abstractmethod
    def format_debug_info_args(self, output_dir_path: str, output_base_name: str) -> list[str]:
        pass

    @abstractmethod
    def format_pch_creation_args(self, pch_header_path: str, pch_output_path: str) -> list[str]:
        pass

    @abstractmethod
    def format_pch_usage_args(self, pch_header_path: str, pch_output_path: str) -> list[str]:
        pass

    @abstractmethod
    def format_definition_compiler_arg(self, name: str, value: Any) -> str:
        pass

    @abstractmethod
    def format_include_dir_arg(self, include_dir: str) -> str:
        pass

    @abstractmethod
    def format_system_include_dir_arg(self, include_dir: str) -> str:
        pass

    @abstractmethod
    def format_settings_args(self, settings: CompilerSettings) -> list[str]:
        pass


def _get_default_compiler_settings(build_type: str) -> CompilerSettings:
    settings = CompilerSettings()
    settings.lang_std = LanguageStandard.CPP17
    settings.lang_conformance = LanguageConformance.MAX

    warn_settings = CompilerWarningSettings()
    warn_settings.level = CompilerWarningLevel.All
    warn_settings.version = "19.33"
    warn_settings.treat_as_errors = True

    # Info: function not inlined
    warn_settings.warning_states["4710"] = CompilerWarningState.DISABLE
    # Info: the compiler performed inlining on the given function, although it was not marked
    # for inlining.
    warn_settings.warning_states["4711"] = CompilerWarningState.DISABLE
    # Compiler will insert Spectre mitigation for memory load if /Qspectre switch specified
    warn_settings.warning_states["5045"] = CompilerWarningState.DISABLE
    # 'struct': '4' bytes padding added after data member 'struct::member'
    warn_settings.warning_states["4820"] = CompilerWarningState.DISABLE

    warn_settings.warning_states["undef"] = CompilerWarningState.DISABLE
    warn_settings.warning_states["reserved-macro-identifier"] = CompilerWarningState.DISABLE
    warn_settings.warning_states["c++98-compat-pedantic"] = CompilerWarningState.DISABLE
    warn_settings.warning_states["newline-eof"] = CompilerWarningState.DISABLE

    settings.warning_settings = warn_settings

    settings.enable_rtti = False
    settings.enable_exceptions = True
    settings.float_behavior = CompilerFloatBehavior.FAST

    if build_type == base.BUILD_TYPE_DEBUG:
        settings.use_debug_runtime = True
        settings.enable_debug_info = True
        settings.optimization_mode = CompilerOptimizationMode.NONE
        settings.enable_security_checks = True
        return settings

    if build_type == base.BUILD_TYPE_RELEASE:
        settings.use_debug_runtime = False
        settings.enable_debug_info = False
        settings.optimization_mode = CompilerOptimizationMode.MAXIMIZE_SPEED
        settings.enable_security_checks = False
        return settings

    if build_type == base.BUILD_TYPE_PROFILING:
        settings = settings
        settings.use_debug_runtime = False
        settings.enable_debug_info = True
        settings.optimization_mode = CompilerOptimizationMode.MAXIMIZE_SPEED
        settings.enable_security_checks = False
        return settings

    raise NotImplementedError()



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any


class WorkspaceInvalidFormat(RuntimeError):
    pass


class TargetKind(Enum):
    STATIC_LIB = auto()
    DYNAMIC_LIB = auto()
    EXECUTABLE = auto()

    @staticmethod
    def from_string(s: str):
        return {
            "rapid_static_lib": TargetKind.STATIC_LIB,
            "rapid_dynamic_lib": TargetKind.DYNAMIC_LIB,
            "rapid_exe": TargetKind.EXECUTABLE,
        }[s]

    def to_string(self):
        return {
            TargetKind.STATIC_LIB: "rapid_static_lib",
            TargetKind.DYNAMIC_LIB: "rapid_dynamic_lib",
            TargetKind.EXECUTABLE: "rapid_exe",
        }[self]


@dataclass
class TargetScopedProperties:
    definitions: dict[str, Any] = field(default_factory=dict)
    dependencies: list[str] = field(default_factory=list)
    externals: list[str] = field(default_factory=list)

    @staticmethod
    def from_dict(d: dict):
        if not isinstance(d, dict):
            raise WorkspaceInvalidFormat()

        if not set(d.keys()).issubset({"definitions", "dependencies", "externals"}):
            raise WorkspaceInvalidFormat()

        definitions = d.get("definitions", {})
        if not isinstance(definitions, dict):
            raise WorkspaceInvalidFormat()

        dependencies = d.get("dependencies", [])
        if not isinstance(dependencies, list):
            raise WorkspaceInvalidFormat()

        externals = d.get("externals", [])
        if not isinstance(externals, list):
            raise WorkspaceInvalidFormat()

        return TargetScopedProperties(definitions=definitions, dependencies=dependencies, externals=externals)


@dataclass
class TargetDefinition:
    name: str
    kind: TargetKind
    private_props: TargetScopedProperties = field(default_factory=TargetScopedProperties)
    public_props: TargetScopedProperties = field(default_factory=TargetScopedProperties)

    @staticmethod
    def from_dict(name: str, d: dict):
        if not isinstance(d, dict):
            raise WorkspaceInvalidFormat()

        if not set(d.keys()).issubset({"kind", "public", "private"}):
            raise WorkspaceInvalidFormat()

        if "kind" not in d.keys():
            raise WorkspaceInvalidFormat()

        try:
            kind = TargetKind.from_string(d["kind"])
        except KeyError:
            raise WorkspaceInvalidFormat()

        public_props = TargetScopedProperties.from_dict(d.get("public", {}))
        private_props = TargetScopedProperties.from_dict(d.get("private", {}))
        return TargetDefinition(name=name, kind=kind, private_props=private_props, public_props=public_props)


@dataclass
class WorkspaceDefinition:
    targets: list[TargetDefinition] = field(default_factory=list)

    def find_target_def(self, name):
        for t in self.targets:
            if t.name == name:
                return t
        return None

    @staticmethod
    def from_dict(d):
        if not isinstance(d, dict):
            raise WorkspaceInvalidFormat()

        if len(d) == 0:
            raise WorkspaceInvalidFormat()

        targets = [TargetDefinition.from_dict(name, v) for name, v in d.items()]
        return WorkspaceDefinition(targets=targets)





from contextlib import contextmanager
import fs
from typing import Any

from .target_graph import (
    TargetImplementation,
)
from .workspace_def import (
    TargetKind,
)
from . import base

from .compiler_base import CompilerArgFormatter, CompilerSettings
from .linker_base import LinkerArgFormatter, LinkerSettings
from .toolchain import (
    CompilerFamily,
    guess_compiler_family_from_compiler_path,
    guess_librarian_family_from_path,
    guess_linker_family_from_path,
    create_compiler_arg_formatter,
    create_librarian_arg_formatter,
    create_linker_arg_formatter,
    get_linker_system_libs,
    get_compiler_system_definitions,
)


class BFFEmitter:
    GENERATED_FILE_HEADER = [
        "/////////////////////////////////////////////////////////////",
        "// This is a generated file, all manual changes will be lost!",
        "/////////////////////////////////////////////////////////////",
    ]
    INDENT_SIZE = 4

    def __init__(self):
        self._lines = []
        self._indent_level = 0

        for line in self.GENERATED_FILE_HEADER:
            self._emit_line(line)

    def _indent(self):
        self._indent_level += 1

    def _dedent(self):
        assert self._indent_level > 0
        self._indent_level -= 1

    def _emit_line(self, content: str):
        if len(content) != 0:
            self._lines.append("{}{}".format(' ' * self._indent_level * self.INDENT_SIZE, content))
        else:
            self._lines.append("")

    def _emit_array(self, name: str, values: list, *, should_split_str=False):
        self._emit_line(".{} = ".format(name))
        self._emit_line("{")
        self._indent()
        for v in values:
            self._emit_line("{},".format(self._format_value(v, should_split_str=should_split_str)))
        self._dedent()
        self._emit_line("}")

    def _emit_struct(self, name: str, dictionary: dict, *, should_split_str=False):
        self._emit_line(".{} = ".format(name))
        self._emit_line("[")
        self._indent()
        for key, value in dictionary.items():
            self.emit_var_assignment(key, value, should_split_str=should_split_str)
        self._dedent()
        self._emit_line("]")

    def _begin_function(self, fn_name: str, args: str):
        self._emit_line("{}('{}')".format(fn_name, args))
        self._emit_line("{")
        self._indent()

    def _end_function(self):
        self._dedent()
        self._emit_line("}")
        self._emit_line("")

    def once(self):
        self._emit_line("#once")

    def include(self, path: str):
        self._emit_line("#include \"{}\"".format(path))

    def line_break(self):
        self._emit_line("")

    def emit_var_assignment(self, name: str, value, *, should_split_str=False):
        if type(value) == list:
            self._emit_array(name, value)
        elif type(value) == dict:
            self._emit_struct(name, value)
        else:
            self._emit_line(".{} = {}".format(name, self._format_value(value, should_split_str=should_split_str)))

    @contextmanager
    def library(self, name: str):
        self._begin_function("Library", name)
        yield
        self._end_function()

    @contextmanager
    def object_list(self, name: str):
        self._begin_function("ObjectList", name)
        yield
        self._end_function()

    @contextmanager
    def dll(self, name: str):
        self._begin_function("DLL", name)
        yield
        self._end_function()

    @contextmanager
    def exe(self, name: str):
        self._begin_function("Executable", name)
        yield
        self._end_function()

    @contextmanager
    def alias(self, name: str):
        self._begin_function("Alias", name)
        yield
        self._end_function()

    @contextmanager
    def copy_dir(self, name: str):
        self._begin_function("CopyDir", name)
        yield
        self._end_function()

    @contextmanager
    def remove_dir(self, name: str):
        self._begin_function("RemoveDir", name)
        yield
        self._end_function()

    @contextmanager
    def compiler(self, name: str):
        self._begin_function("Compiler", name)
        yield
        self._end_function()

    @contextmanager
    def vcxproject(self, name: str):
        self._begin_function("VCXProject", name)
        yield
        self._end_function()

    @contextmanager
    def vssolution(self, name: str):
        self._begin_function("VSSolution", name)
        yield
        self._end_function()

    def build(self):
        return "\n".join(self._lines)

    def _format_value(self, value, *, should_split_str):

        def format_default(v):
            return "{}".format(v)

        def str_smart_split(s: str):
            splits = []
            inside_quoted = False
            delim = " "
            curr = ""
            for c in s:
                if not inside_quoted:
                    if c == "\"":
                        inside_quoted = True
                        curr += c
                    elif c == delim:
                        splits.append(curr)
                        curr = delim
                    else:
                        curr += c
                else:
                    if c == "\"":
                        inside_quoted = False
                        curr += c
                    else:
                        curr += c
            if len(curr) != 0:
                splits.append(curr)
            return splits

        def format_string(s: str):
            if len(s) == 0:
                return "''"

            if not should_split_str:
                return f"'{s}'"

            lines = []
            col_limit = 80

            if len(s) > col_limit:
                lines = str_smart_split(s)
            else:
                lines.append(s)

            result = ""
            result += "'{}'".format(lines[0])
            if len(lines) > 1:
                for line in lines[1:]:
                    result += f"\n{' ' * (self._indent_level + 1) * self.INDENT_SIZE}+ '{line}'"
            return result

        def format_bool(b: bool):
            return "true" if b else "false"

        formatters = {
            str: format_string,
            bool: format_bool,
        }

        f = formatters.get(type(value), format_default)
        return f(value)


def get_target_kind_extension(target_kind):
    if target_kind == TargetKind.STATIC_LIB:
        return "lib"
    elif target_kind == TargetKind.DYNAMIC_LIB:
        return "dll"
    elif target_kind == TargetKind.EXECUTABLE:
        return "exe"
    else:
        raise NotImplementedError("Target kind of this type is not implemented.")


def _build_compiler_include_dir_args(
        arg_formatter: CompilerArgFormatter,
        system_include_dirs: list[str],
        target_include_dirs: list[str],
):
    args = []
    args += [arg_formatter.format_system_include_dir_arg(include_dir) for include_dir in system_include_dirs]
    args += [arg_formatter.format_include_dir_arg(include_dir) for include_dir in target_include_dirs]
    return args


def _build_compiler_definitions_args(
        arg_formatter: CompilerArgFormatter,
        system_definitions: dict[str, Any],
        target_definitions: dict[str, Any],
):
    args = []
    args += [arg_formatter.format_definition_compiler_arg(name, value) for name, value in system_definitions.items()]
    args += [arg_formatter.format_definition_compiler_arg(name, value) for name, value in target_definitions.items()]
    return args


def _build_compiler_compiland_args(
        arg_formatter: CompilerArgFormatter,
        output_dir_path: str, output_base_name: str,
        pch_header_path: str | None, pch_output_path: str | None,
):
    args = []
    args += arg_formatter.format_beginning_args()
    args += arg_formatter.format_object_compilation_args()
    if pch_header_path is not None and pch_output_path is not None:
        args += arg_formatter.format_pch_usage_args(pch_header_path, pch_output_path)
    args += arg_formatter.format_debug_info_args(output_dir_path, output_base_name)
    return args


def _build_compiler_pch_args(
        arg_formatter: CompilerArgFormatter,
        output_dir_path: str, output_base_name: str,
        pch_header_path: str, pch_output_path: str,
):
    args = []
    args += arg_formatter.format_beginning_args()
    args += arg_formatter.format_pch_creation_args(pch_header_path, pch_output_path)
    args += arg_formatter.format_debug_info_args(output_dir_path, output_base_name)
    return args


def _build_linker_args(
        arg_formatter: LinkerArgFormatter,
        output_dir_path: str, output_base_name: str, output_target_kind: TargetKind,
        settings: LinkerSettings,
        base_link_libs: list[str], base_lib_dirs: list[str],
        external_link_libs: list[str], external_lib_dirs: list[str]
):
    if output_target_kind != TargetKind.DYNAMIC_LIB and output_target_kind != TargetKind.EXECUTABLE:
        raise RuntimeError(f"Target kind '{output_target_kind.name}' is not created using linker.")

    args = arg_formatter.format_input_output_declaration(output_dir_path, output_base_name, output_target_kind)
    args += arg_formatter.format_settings_args(settings)

    args += [arg_formatter.format_lib_arg(lib) for lib in base_link_libs]
    args += [arg_formatter.format_lib_dir_arg(lib_dir) for lib_dir in base_lib_dirs]

    # External libs are added to commandl ine manully, not through "Libraries" variable.
    # If the extension is skipped, fastbuild tries to use "obj"
    args += [arg_formatter.format_lib_arg(f"{lib}.{get_target_kind_extension(TargetKind.STATIC_LIB)}")
             for lib in external_link_libs]
    args += [arg_formatter.format_lib_dir_arg(lib_dir) for lib_dir in external_lib_dirs]

    return args


class BuildScriptEmitterFastbuild(base.BuildScriptEmitter):
    ALL_TARGETS_TARGET_NAME = "rapid_all"
    SLN_TARGET_NAME = "rapid_sln"
    CLEAN_TARGET_NAME = "rapid_clean"
    TARGET_WORKING_DIR_NAME = "data"

    HEADER_EXTENSIONS = ["h", "hh", "hxx", "hpp"]
    SOURCE_EXTENSIONS = ["c", "cc", "cxx", "cpp"]

    PCH_OUTPUT_EXT = "pch"

    def filename(self):
        return "fbuild.bff"

    def contents(self, wks_name, source_dir_abs_path, configure_dir_abs_path, build_dir_abs_path,
                 toolchains_manifest,
                 toolchains_settings_provider: base.ToolchainSettingsProvider,
                 build_types,
                 targets_names,
                 targets_impls
                 ) -> str:

        assert (set(toolchains_manifest.keys()).issubset(toolchains_settings_provider.get_toolchains()))

        intermediate_dir_abs_path = fs.path.join(build_dir_abs_path, "intermediate")
        deploy_dir_abs_path = fs.path.join(build_dir_abs_path, "bin")

        out_bff = BFFEmitter()
        out_bff.once()

        self._emit_compilers(out_bff, toolchains_manifest)
        self._emit_compilation_targets(out_bff, intermediate_dir_abs_path, deploy_dir_abs_path, toolchains_manifest,
                                       toolchains_settings_provider, build_types, targets_names, targets_impls)
        self._emit_vs_solution_targets(out_bff, wks_name, source_dir_abs_path, configure_dir_abs_path, deploy_dir_abs_path,
                                       toolchains_manifest, toolchains_settings_provider, build_types, targets_names, targets_impls)

        with out_bff.remove_dir(self.CLEAN_TARGET_NAME):
            out_bff.emit_var_assignment("RemovePaths", build_dir_abs_path)
            out_bff.emit_var_assignment("RemovePathsRecurse", True)
            
        return out_bff.build()

    @classmethod
    def _emit_compilers(cls, out_bff: BFFEmitter, toolchains_manifest):
        
        for toolchain_name, toolchain_def in toolchains_manifest.items():
            with out_bff.compiler(cls._format_compiler_node_name(toolchain_name)):
                compiler_path = toolchain_def["compiler_path"]
                compiler_family = guess_compiler_family_from_compiler_path(compiler_path)

                out_bff.emit_var_assignment("Executable", compiler_path)
                out_bff.emit_var_assignment("CompilerFamily", cls._compiler_family_to_fastbuild_value(compiler_family))
                out_bff.emit_var_assignment("ExtraFiles", toolchain_def["compiler_extra_files"])

    @classmethod
    def _search_pch_files_in_dir(cls, path: str):
        with fs.open_fs(path) as input_fs:
            pch_header_patterns = [f"*_pch.{ext}" for ext in cls.HEADER_EXTENSIONS]
            pch_source_patterns = [f"*_pch.{ext}" for ext in cls.SOURCE_EXTENSIONS]

            pch_headers = list(input_fs.filterdir(".", files=pch_header_patterns))
            pch_sources = list(input_fs.filterdir(".", files=pch_source_patterns))

            # For some reason dirs are also found using filterdir()
            pch_headers = [f for f in pch_headers if f.is_file]
            pch_sources = [f for f in pch_sources if f.is_file]

            if len(pch_headers) == 0 and len(pch_sources) == 0:
                return None, None
            elif len(pch_headers) == 1 and len(pch_sources) == 1:
                return pch_headers[0].make_path(path), pch_sources[0].make_path(path)
            else:
                msg = f"Source dir \"{path}\" should contain a single PCH header and a single PCH source.\n"
                
                msg += "When searching for PCH header using patterns: "
                msg += " ".join(pch_header_patterns)
                msg += "\n"
                if len(pch_headers) == 0:
                    msg += f"found 0 files.\n"
                else:
                    msg += f"found {len(pch_headers)} files:\n"
                    for f in pch_headers:
                        msg += f"* {f.name}\n"

                msg += "When searching for PCH source using patterns: "
                msg += " ".join(pch_source_patterns)
                msg += "\n"
                if len(pch_sources) == 0:
                    msg += f"found 0 files.\n"
                else:
                    msg += f"found {len(pch_sources)} files:\n"
                    for f in pch_sources:
                        msg += f"* {f.name}\n"
                    
                raise RuntimeError(msg)

    def _emit_compilation_targets(
            self,
            out_bff: BFFEmitter,
            intermediate_dir_abs_path: str,
            deploy_dir_abs_path: str,
            toolchains_manifest: dict[str, Any],
            toolchains_settings_provider: base.ToolchainSettingsProvider,
            build_types: list[str],
            targets_names: list[str],
            targets_impls: dict[str, TargetImplementation]
    ):

        all_nodes: list[str] = []
        per_toolchain_nodes: dict[str, list[str]] = {
            toolchain_name: [] for toolchain_name in toolchains_manifest.keys()
        }
        per_build_type_nodes: dict[str, list[str]] = {
            build_type: [] for build_type in build_types
        }
        config_vector_nodes: dict[str, dict[str, list[str]]] = {
            toolchain_name: {
                build_type: [] for build_type in build_types
            } for toolchain_name in toolchains_manifest.keys()
        }
        external_deploy_nodes = set()

        for toolchain_name, toolchain_def in toolchains_manifest.items():

            toolchain_compiler_family = guess_compiler_family_from_compiler_path(toolchain_def["compiler_path"])
            toolchain_librarian_family = guess_librarian_family_from_path(toolchain_def["librarian_path"])
            toolchain_linker_family = guess_linker_family_from_path(toolchain_def["linker_path"])

            compiler_arg_formatter = create_compiler_arg_formatter(toolchain_compiler_family)
            librarian_arg_formatter = create_librarian_arg_formatter(toolchain_librarian_family)
            linker_arg_formatter = create_linker_arg_formatter(toolchain_linker_family)

            toolchain_include_dirs = toolchain_def["toolchain_include_dirs"]
            toolchain_lib_dirs = toolchain_def["toolchain_lib_dirs"]

            toolchain_settings = toolchains_settings_provider.get_toolchain_settings(toolchain_name)

            for build_type in build_types:

                toolchain_build_type_compiler_settings = toolchain_settings.get_compiler_settings(build_type)
                toolchain_build_type_definitions = get_compiler_system_definitions(toolchain_compiler_family, build_type)
                # TODO: Validate that user-provided compiler definitions do not override system definitions.
                toolchain_build_type_definitions |= toolchain_settings.get_compiler_definitions(build_type)

                toolchain_build_type_linker_settings = toolchain_settings.get_linker_settings(build_type)
                toolchain_build_type_link_libs = get_linker_system_libs(toolchain_linker_family, build_type)

                build_type_deploy_dir_path = fs.path.join(deploy_dir_abs_path, build_type)

                for target_name in targets_names:

                    target_impl = targets_impls[base.build_target_key(target_name, toolchain_name, build_type)]

                    target_node_name = self._format_target_node_name(target_name, toolchain_name, build_type)
                    make_node_name = self._format_make_node_name(target_node_name)
                    deploy_node_name = self._format_deploy_node_name(target_node_name)

                    target_kind = target_impl.kind
                    target_output_dir_path = fs.path.join(intermediate_dir_abs_path, target_node_name)

                    def emit_bff_compiler_args():
                        out_bff.emit_var_assignment("Compiler", self._format_compiler_node_name(toolchain_name))

                        compiler_input_dir = target_impl.source_dir
                        out_bff.emit_var_assignment("CompilerInputPath", compiler_input_dir)

                        out_bff.emit_var_assignment("CompilerInputPathRecurse", True)

                        compiler_input_pattern = [f"*.{ext}" for ext in self.SOURCE_EXTENSIONS]
                        out_bff.emit_var_assignment("CompilerInputPattern", compiler_input_pattern)

                        compiler_output_dir = fs.path.join(target_output_dir_path, target_name)
                        out_bff.emit_var_assignment("CompilerOutputPath", compiler_output_dir)

                        # TODO: Per-target compiler settings.
                        per_target_compiler_settings = CompilerSettings()
                        def combine(lhs, rhs):
                            return lhs

                        target_compiler_settings = combine(toolchain_build_type_compiler_settings,
                                                           per_target_compiler_settings)

                        common_args = []
                        common_args += compiler_arg_formatter.format_settings_args(target_compiler_settings)
                        common_args += _build_compiler_include_dir_args(compiler_arg_formatter,
                                                                        toolchain_include_dirs,
                                                                        target_impl.include_dirs)
                        common_args += _build_compiler_definitions_args(compiler_arg_formatter,
                                                                        toolchain_build_type_definitions,
                                                                        target_impl.definitions)


                        pch_header_file, pch_source_file = self._search_pch_files_in_dir(compiler_input_dir)
                        pch_output_file = fs.path.join(target_output_dir_path,
                                                       f"{target_node_name}.{self.PCH_OUTPUT_EXT}")

                        compiland_args = _build_compiler_compiland_args(
                            compiler_arg_formatter,
                            target_output_dir_path, target_node_name,
                            pch_header_file, pch_output_file
                        )
                        compiland_args += common_args


                        out_bff.emit_var_assignment("CompilerOptions", " ".join(compiland_args),
                                                    should_split_str=True)

                        if pch_header_file is not None and pch_source_file is not None:
                            out_bff.line_break()

                            out_bff.emit_var_assignment("PCHInputFile", pch_source_file)
                            out_bff.emit_var_assignment("PCHOutputFile", pch_output_file)

                            pch_args = _build_compiler_pch_args(
                                compiler_arg_formatter,
                                target_output_dir_path, target_node_name,
                                pch_header_file, pch_output_file
                            )
                            pch_args += common_args

                            out_bff.emit_var_assignment("PCHOptions", " ".join(pch_args), should_split_str=True)
                            out_bff.line_break()

                    if target_kind == TargetKind.STATIC_LIB:

                        def emit_bff_librarian_args():
                            librarian_output_file = fs.path.join(target_output_dir_path,
                                                                 self._format_target_filename(target_node_name,
                                                                                              target_kind))

                            out_bff.emit_var_assignment("Librarian", toolchain_def["librarian_path"])
                            out_bff.emit_var_assignment("LibrarianOutput", librarian_output_file)
                            out_bff.emit_var_assignment("LibrarianOptions",
                                                        " ".join(librarian_arg_formatter.format_input_output_args()),
                                                        should_split_str=True)

                        with out_bff.library(make_node_name):
                            out_bff.emit_var_assignment("Hidden", True)
                            emit_bff_compiler_args()
                            emit_bff_librarian_args()

                        # Deploy step does nothing for static libs.
                        with out_bff.alias(deploy_node_name):
                            out_bff.emit_var_assignment("Hidden", True)
                            out_bff.emit_var_assignment("Targets", make_node_name)

                        with out_bff.alias(target_node_name):
                            out_bff.emit_var_assignment("Targets", deploy_node_name)

                    elif target_kind == TargetKind.DYNAMIC_LIB or target_kind == TargetKind.EXECUTABLE:

                        object_list_node_name = self._format_obj_node_name(target_node_name)
                        with out_bff.object_list(object_list_node_name):
                            out_bff.emit_var_assignment("Hidden", True)
                            emit_bff_compiler_args()

                        def emit_bff_linker_args():
                            linker_output_file = fs.path.join(target_output_dir_path,
                                                              self._format_target_filename(target_node_name,
                                                                                           target_kind))

                            libraries = [object_list_node_name]
                            libraries += [self._format_make_node_name(
                                self._format_target_node_name(lib, toolchain_name, build_type))
                                for lib in target_impl.link_libs]

                            linker_options = _build_linker_args(
                                linker_arg_formatter, target_output_dir_path, target_node_name, target_kind,
                                toolchain_build_type_linker_settings,
                                toolchain_build_type_link_libs, toolchain_lib_dirs,
                                target_impl.link_libs_external, target_impl.link_libs_external_dirs
                            )

                            out_bff.emit_var_assignment("Libraries", libraries)
                            out_bff.emit_var_assignment("Linker", toolchain_def["linker_path"])
                            out_bff.emit_var_assignment("LinkerOutput", linker_output_file)
                            out_bff.emit_var_assignment("LinkerOptions", " ".join(linker_options),
                                                        should_split_str=True)

                        if target_kind == TargetKind.DYNAMIC_LIB:
                            with out_bff.dll(make_node_name):
                                # out_bff.emit_var_assignment("Hidden", True)  # DLL does not support .Hidden
                                emit_bff_linker_args()

                        elif target_kind == TargetKind.EXECUTABLE:
                            with out_bff.exe(make_node_name):
                                # out_bff.emit_var_assignment("Hidden", True)  # Executable does not support .Hidden
                                emit_bff_linker_args()

                        deploy_deps = [make_node_name]
                        deploy_deps += [
                            self._format_deploy_node_name(self._format_target_node_name(lib, toolchain_name, build_type))
                            for lib in target_impl.load_time_libs]

                        for path in target_impl.load_time_libs_external:
                            deploy_external = self._format_deploy_node_name(path)
                            if path not in external_deploy_nodes:
                                external_deploy_nodes.add(path)

                                with out_bff.copy_dir(deploy_external):
                                    out_bff.emit_var_assignment("Hidden", True)
                                    out_bff.emit_var_assignment("PreBuildDependencies", [])
                                    out_bff.emit_var_assignment("SourcePaths", fs.path.dirname(path))
                                    out_bff.emit_var_assignment("SourcePathsPattern", [(fs.path.basename(path))])
                                    out_bff.emit_var_assignment("Dest", build_type_deploy_dir_path)
                            deploy_deps.append(deploy_external)

                        with out_bff.copy_dir(deploy_node_name):
                            out_bff.emit_var_assignment("Hidden", True)
                            out_bff.emit_var_assignment("PreBuildDependencies", deploy_deps)
                            out_bff.emit_var_assignment("SourcePaths", target_output_dir_path)
                            out_bff.emit_var_assignment("SourcePathsPattern",
                                                        [self._format_target_glob_pattern(target_kind), "*.pdb"])
                            out_bff.emit_var_assignment("Dest", build_type_deploy_dir_path)

                        with out_bff.alias(target_node_name):
                            out_bff.emit_var_assignment("Targets", deploy_node_name)

                    else:
                        raise NotImplementedError("unsupported target kind")

                    all_nodes.append(target_node_name)
                    per_build_type_nodes[build_type].append(target_node_name)
                    per_toolchain_nodes[toolchain_name].append(target_node_name)
                    config_vector_nodes[toolchain_name][build_type].append(target_node_name)

        for build_type, nodes in per_build_type_nodes.items():
            with out_bff.alias(f"{self.ALL_TARGETS_TARGET_NAME}-{build_type}"):
                out_bff.emit_var_assignment("Targets", nodes)

        for toolchain_name, nodes in per_toolchain_nodes.items():
            with out_bff.alias(f"{self.ALL_TARGETS_TARGET_NAME}-{toolchain_name}"):
                out_bff.emit_var_assignment("Targets", nodes)

        for toolchain_name, per_toolchain_per_build_type_visible_nodes in config_vector_nodes.items():
            for build_type, nodes in per_toolchain_per_build_type_visible_nodes.items():
                with out_bff.alias(f"{self.ALL_TARGETS_TARGET_NAME}-{toolchain_name}-{build_type}"):
                    out_bff.emit_var_assignment("Targets", nodes)

        with out_bff.alias(f"{self.ALL_TARGETS_TARGET_NAME}"):
            out_bff.emit_var_assignment("Targets", all_nodes)

    # noinspection PyUnreachableCode
    def _emit_vs_solution_targets(
            self,
            out_bff: BFFEmitter,
            wks_name: str,
            source_dir_abs_path: str,
            configure_dir_abs_path: str,
            deploy_dir_abs_path: str,
            toolchains_manifest: dict[str, Any],
            toolchains_settings_provider: base.ToolchainSettingsProvider,
            build_types: list[str],
            targets_names: list[str],
            targets_impls: dict[str, TargetImplementation]
    ):

        bff_path = fs.path.join(f"{configure_dir_abs_path}", self.filename())
        vs_dir_abs_path = fs.path.join(configure_dir_abs_path, "vs")

        latest_msvc_toolset = None
        latest_vs_version = None
        if base.TOOLCHAIN_VS17_MSVC in toolchains_manifest.keys():
            latest_msvc_toolset = "v143"
            latest_vs_version = "17.0"
        elif base.TOOLCHAIN_VS16_MSVC in toolchains_manifest.keys():
            latest_msvc_toolset = "v142"
            latest_vs_version = "16.0"
        else:
            return

        lang_standard = "c++20"

        # Fastbuild can deduce these automatically based of "Target", manual spec overrides it.
        # Currently, our "Target" is pointing to an alias, which points to a deployment step,
        # and fastbuild is unable to deduce those.
        explicitly_define_intellisense_args = True

        project_nodes = []
        for target_name in targets_names:
            target_source_dir_abs_path = fs.path.join(source_dir_abs_path, target_name)
            project_out_abs_path = fs.path.join(vs_dir_abs_path, f"{target_name}.vcxproj")

            project_node_name = f"vcxproj-{target_name}"
            with out_bff.vcxproject(project_node_name):
                out_bff.emit_var_assignment("ProjectOutput", project_out_abs_path)
                out_bff.emit_var_assignment("ProjectInputPaths", target_source_dir_abs_path)
                out_bff.emit_var_assignment("ProjectBasePath", target_source_dir_abs_path)

                out_bff._emit_line(".ProjectConfigs = {}")

                for toolchain_name, toolchain_def in toolchains_manifest.items():

                    toolchain_include_dirs = toolchain_def["toolchain_include_dirs"]

                    toolchain_settings = toolchains_settings_provider.get_toolchain_settings(toolchain_name)

                    for build_type in build_types:

                        build_type_deploy_dir_path = fs.path.join(deploy_dir_abs_path, build_type)

                        toolchain_build_type_definitions = get_compiler_system_definitions(CompilerFamily.MSVC,
                                                                                           build_type)
                        # TODO: Validate that user-provided compiler definitions do not override system definitions.
                        toolchain_build_type_definitions |= toolchain_settings.get_compiler_definitions(build_type)

                        target_impl = targets_impls[base.build_target_key(target_name, toolchain_name, build_type)]
                        target_node_name = self._format_target_node_name(target_name, toolchain_name, build_type)

                        out_bff._emit_line(".cfg = [")
                        out_bff._indent()

                        out_bff.emit_var_assignment("Platform", "x64")
                        out_bff.emit_var_assignment("PlatformToolset", latest_msvc_toolset)
                        out_bff.emit_var_assignment("Config", f"{toolchain_name}-{build_type}")
                        # out_bff.emit_var_assignment("Config", "Debug")
                        out_bff.line_break()
                        out_bff.emit_var_assignment("Target", target_node_name)
                        out_bff.emit_var_assignment("ProjectBuildCommand",
                                                    f"fbuild -ide -config \"{bff_path}\" {target_node_name}")
                        out_bff.emit_var_assignment("ProjectRebuildCommand",
                                                    f"fbuild -ide -config \"{bff_path}\" -clean {target_node_name}")
                        # ProjectCleanCommand
                        out_bff.emit_var_assignment("LocalDebuggerCommand",
                                                    fs.path.join(build_type_deploy_dir_path, f"{target_node_name}.exe"))
                        out_bff.emit_var_assignment("LocalDebuggerWorkingDirectory",
                                                    fs.path.join(source_dir_abs_path, target_name, self.TARGET_WORKING_DIR_NAME))

                        out_bff.line_break()

                        if explicitly_define_intellisense_args:
                            include_dirs_intellisense = ""

                            for include_dir in toolchain_include_dirs:
                                include_dirs_intellisense += f"{include_dir};"

                            for include_dir in target_impl.include_dirs:
                                include_dirs_intellisense += f"{include_dir};"

                            out_bff.emit_var_assignment("IncludeSearchPath", include_dirs_intellisense)

                            definitions_intellisense = ""

                            for name, value in toolchain_build_type_definitions.items():
                                if value is not None:
                                    definitions_intellisense += f"{name}={value};"
                                else:
                                    definitions_intellisense += f"{name};"

                            for name, value in target_impl.definitions.items():
                                if value is not None:
                                    definitions_intellisense += f"{name}={value};"
                                else:
                                    definitions_intellisense += f"{name};"

                            out_bff.emit_var_assignment("PreprocessorDefinitions", definitions_intellisense)

                            # TODO: generate AdditionalOptions from compiler options - standard and warning settings
                            out_bff.emit_var_assignment("AdditionalOptions", f"/std:{lang_standard}")

                        out_bff._dedent()
                        out_bff._emit_line("]")

                        out_bff._emit_line(".ProjectConfigs + .cfg")

            project_nodes.append(project_node_name)

        # special 'all' project
        all_targets_project_path = fs.path.join(vs_dir_abs_path, f"{self.ALL_TARGETS_TARGET_NAME}.vcxproj")

        project_node_name = f"vcxproj-{self.ALL_TARGETS_TARGET_NAME}"
        with out_bff.vcxproject(project_node_name):
            out_bff.emit_var_assignment("ProjectOutput", all_targets_project_path)

            out_bff._emit_line(".ProjectConfigs = {}")

            for toolchain_name in toolchains_manifest.keys():
                for build_type in build_types:
                    target_node_name = self._format_target_node_name(self.ALL_TARGETS_TARGET_NAME, toolchain_name,
                                                                     build_type)

                    out_bff._emit_line(".cfg = [")
                    out_bff._indent()

                    out_bff.emit_var_assignment("Platform", "x64")
                    out_bff.emit_var_assignment("PlatformToolset", latest_msvc_toolset)
                    out_bff.emit_var_assignment("Config", f"{toolchain_name}-{build_type}")
                    # out_bff.emit_var_assignment("Config", "Debug")
                    out_bff.line_break()
                    out_bff.emit_var_assignment("Target", target_node_name)
                    out_bff.emit_var_assignment("ProjectBuildCommand",
                                                f"fbuild -ide -config \"{bff_path}\" {target_node_name}")
                    out_bff.emit_var_assignment("ProjectRebuildCommand",
                                                f"fbuild -ide -config \"{bff_path}\" -clean {target_node_name}")
                    # ProjectCleanCommand
                    out_bff.emit_var_assignment("LocalDebuggerCommand", "")

                    out_bff.emit_var_assignment("AdditionalOptions", f"/std:{lang_standard}")

                    out_bff._dedent()
                    out_bff._emit_line("]")
                    out_bff._emit_line(".ProjectConfigs + .cfg")

        project_nodes.append(project_node_name)

        # solution
        sln_path = fs.path.join(vs_dir_abs_path, f"{wks_name}.sln")

        with out_bff.vssolution(self.SLN_TARGET_NAME):
            out_bff.emit_var_assignment("SolutionOutput", sln_path)
            out_bff.emit_var_assignment("SolutionProjects", project_nodes)
            out_bff.line_break()

            if False:
                sln_configs = []

                for toolchain_name in toolchains_manifest.keys():
                    for build_type in build_types:
                        cfg = {
                            "Platform": "x64",
                            "Config": f"{toolchain_name}-{build_type}",
                            "SolutionPlatform": toolchain_name,
                            "SolutionConfig": build_type,
                            "SolutionBuildProject": all_targets_project_path
                        }
                        sln_configs.append(cfg)
                out_bff.emit_var_assignment("SolutionConfigs", sln_configs)
                out_bff.line_break()

                sln_deps = [
                    {
                        "Projects": project_nodes,
                        "Dependencies": [all_targets_project_path]
                    }
                ]

                out_bff.emit_var_assignment("SolutionDependencies", sln_deps, should_split_str=True)

            else:
                out_bff._emit_line(".SolutionConfigs = {}")

                for toolchain_name in toolchains_manifest.keys():
                    for build_type in build_types:
                        out_bff._emit_line(".cfg = [")
                        out_bff._indent()

                        out_bff.emit_var_assignment("Platform", "x64")
                        out_bff.emit_var_assignment("Config", f"{toolchain_name}-{build_type}")
                        out_bff.emit_var_assignment("SolutionPlatform", toolchain_name)
                        out_bff.emit_var_assignment("SolutionConfig", build_type)
                        out_bff.emit_var_assignment("SolutionBuildProject", all_targets_project_path)

                        out_bff._dedent()
                        out_bff._emit_line("]")
                        out_bff._emit_line(".SolutionConfigs + .cfg")
                out_bff.line_break()

                out_bff._emit_line(".SolutionDependencies = {}")

                out_bff._emit_line(".dep = [")
                out_bff._indent()

                out_bff.emit_var_assignment("Projects", project_nodes)
                out_bff.emit_var_assignment("Dependencies", [all_targets_project_path])

                out_bff._dedent()
                out_bff._emit_line("]")
                out_bff._emit_line(".SolutionDependencies + .dep")
            out_bff.line_break()

            out_bff.emit_var_assignment("SolutionVisualStudioVersion", latest_vs_version)
            out_bff.emit_var_assignment("SolutionMinimumVisualStudioVersion", latest_vs_version)

    @staticmethod
    def _format_target_filename(target_name, target_kind):
        ext = get_target_kind_extension(target_kind)
        if len(ext) > 0:
            return f"{target_name}.{ext}"
        else:
            return target_name

    @classmethod
    def _format_target_glob_pattern(cls, target_kind):
        return cls._format_target_filename("*", target_kind)

    @staticmethod
    def _format_compiler_node_name(toolchain_name: str) -> str:
        return f"compiler-{toolchain_name}"

    @staticmethod
    def _format_target_node_name(target_name: str, toolchain_name: str, build_type_name: str) -> str:
        return f"{target_name}-{toolchain_name}-{build_type_name}"

    @staticmethod
    def _format_obj_node_name(target_node_name: str):
        return f"obj-{target_node_name}"

    @staticmethod
    def _format_make_node_name(target_node_name: str):
        return f"make-{target_node_name}"

    @staticmethod
    def _format_deploy_node_name(target_node_name: str):
        return f"deploy-{target_node_name}"

    @staticmethod
    def _compiler_family_to_fastbuild_value(f: CompilerFamily):
        try:
            return {
                CompilerFamily.MSVC: "msvc",
                CompilerFamily.CLANG: "clang",
                CompilerFamily.CLANG_CL: "clang-cl",
                CompilerFamily.GCC: "gcc"
            }[f]
        except KeyError:
            raise NotImplementedError(f"Missing conversion of compiler family '{f.name}' to fastbuild known value.")







__all__ = [
    "base",

    "compiler_base",
    "compiler_llvm",
    "compiler_msvc",

    "librarian_base",
    "librarian_llvm",
    "librarian_msvc",

    "linker_base",
    "linker_llvm",
    "linker_msvc",

    "toolchain",

    "conan",
    "fastbuild",

    "graph",
    "target_graph",

    "workspace_def",
    "workspace",
]


from . import base
from .compiler_base import (CompilerArgFormatter, CompilerSettings, LanguageStandard, UnsupportedLanguageError,
    LanguageConformance, CompilerWarningLevel, CompilerWarningState, CompilerWarningSettings, CompilerFloatBehavior,
    CompilerOptimizationMode, UnsupportedSettingError
)


class CompilerArgFormatterClang(CompilerArgFormatter):
    def format_beginning_args(self) -> list[str]:
        return []

    def format_object_compilation_args(self) -> list[str]:
        return ["-c", "%1", "-o", "%2"]

    def format_debug_info_args(self, output_dir_path: str, output_base_name: str) -> list[str]:
        return []

    def format_pch_creation_args(self, pch_header_path: str, pch_output_path: str) -> list[str]:
        return [
            "-x", "c++-header", f"\"{pch_header_path}\"",
            "-o", "\"%2\""  # PCH output file (compiled .pch)
        ]

    def format_pch_usage_args(self, pch_header_path: str, pch_output_path: str) -> list[str]:
        return [
            "-include-pch", f"\"{pch_output_path}\""
        ]

    def format_definition_compiler_arg(self, name: str, value: base.Any) -> str:
        if value is not None:
            return f"-D\"{name}={value}\""
        else:
            return f"-D\"{name}\""

    def format_include_dir_arg(self, include_dir: str) -> str:
        return f"-I\"{include_dir}\""

    def format_system_include_dir_arg(self, include_dir: str) -> str:
        return f"-isystem\"{include_dir}\""

    @staticmethod
    def _format_language_std(std: LanguageStandard) -> list[str]:
        match std:
            case LanguageStandard.C89:
                return ['-x', 'c', '-std=c89']
            case LanguageStandard.C99:
                return ['-x', 'c', '-std=c99']
            case LanguageStandard.C11:
                return ['-x', 'c', '-std=c11']
            case LanguageStandard.C17:
                return ['-x', 'c', '-std=c17']
            case LanguageStandard.C23:
                return ['-x', 'c', '-std=c23']
            case LanguageStandard.C_LATEST:
                return ['-x', 'c', '-std=c23']
            case LanguageStandard.CPP03:
                return ['-x', 'c++', '-std=c++98']
            case LanguageStandard.CPP11:
                return ['-x', 'c++', '-std=c++11']
            case LanguageStandard.CPP14:
                return ['-x', 'c++', '-std=c++14']
            case LanguageStandard.CPP17:
                return ['-x', 'c++', '-std=c++17']
            case LanguageStandard.CPP20:
                return ['-x', 'c++', '-std=c++20']
            case LanguageStandard.CPP23:
                return ['-x', 'c++', '-std=c++23']
            case LanguageStandard.CPP_LATEST:
                return ['-x', 'c++', '-std=c++2c']
            case _:
                raise UnsupportedLanguageError(f"Unhandled language standard '{std}'.")

    @staticmethod
    def _format_language_conformance(lang_conformance: LanguageConformance) -> list[str]:
        match lang_conformance:
            case LanguageConformance.MIN:
                return [
                    "--no-pedantic",
                    "-fms-extensions",
                    # "-fms-volatile" #TODO: Unknown argument
                ]
            case LanguageConformance.MAX:
                return [
                    "--pedantic",
                    "--pedantic-errors",
                    "-fno-ms-extensions",
                    # "-fno-ms-volatile" #TODO: Unknown argument
                ]
            case _:
                raise UnsupportedLanguageError(f"Unhandled language conformance option '{lang_conformance}'.")

    @staticmethod
    def _format_warning_level(warning_level: CompilerWarningLevel) -> list[str]:
        match warning_level:
            case CompilerWarningLevel.W0:
                return ["-w"]
            case CompilerWarningLevel.W1:
                return ["-Wall"]
            case CompilerWarningLevel.W2:
                return ["-Wall"]
            case CompilerWarningLevel.W3:
                return ["-Wall"]
            case CompilerWarningLevel.W4:
                return ["-Wall", "-Wextra"]
            case CompilerWarningLevel.All:
                return ["-Wall", "-Wextra", "-Weverything"]
            case _:
                raise UnsupportedSettingError()

    @staticmethod
    def _format_warning_version(ver: str) -> list[str]:
        return []

    @staticmethod
    def _format_warning_state(warn: str, state: CompilerWarningState) -> list[str]:
        match state:
            case CompilerWarningState.DISABLE:
                return [f"-Wno-{warn}"]
            case CompilerWarningState.ENABLE:
                return [f"-W{warn}"]
            case CompilerWarningState.REPORT_ONCE:
                return [f"-W{warn}"]
            case CompilerWarningState.TREAT_AS_ERROR:
                return [f"-Werror={warn}"]
            case _:
                raise UnsupportedSettingError()

    @classmethod
    def _is_known_warning(cls, warn):
        return any(c.isalpha() for c in warn)

    @classmethod
    def _format_warning_states(cls, states: dict[str, CompilerWarningState]) -> list[str]:
        l = []
        for warn, state in states.items():
            if cls._is_known_warning(warn):
                l += cls._format_warning_state(warn, state)
        return l

    @classmethod
    def _format_warning_settings(cls, warning_settings: CompilerWarningSettings) -> list[str]:
        l = []
        l += cls._format_warning_level(warning_settings.level)
        l += cls._format_warning_version(warning_settings.version)
        if warning_settings.treat_as_errors:
            l += ["-Werror"]
        l += cls._format_warning_states(warning_settings.warning_states)
        l += ["-Wno-system-headers"]
        return l

    @staticmethod
    def _format_float_behavior(beh: CompilerFloatBehavior) -> list[str]:
        match beh:
            case CompilerFloatBehavior.PRECISE:
                return ["-ffp-model=precise"]
            case CompilerFloatBehavior.STRICT:
                return ["-ffp-model=strict"]
            case CompilerFloatBehavior.FAST:
                return ["-ffp-model=fast"]
            case _:
                raise UnsupportedSettingError()

    @staticmethod
    def _format_optimization_mode(opt: CompilerOptimizationMode) -> list[str]:
        match opt:
            case CompilerOptimizationMode.NONE:
                return ["-O0"]
            case CompilerOptimizationMode.MINIMIZE_SIZE:
                return ["-O3", "-Osize"]
            case CompilerOptimizationMode.MAXIMIZE_SPEED:
                return ["-O3", "-Ofast"]

    @classmethod
    def _format_compiler_settings(cls, settings: CompilerSettings) -> list[str]:
        l = []
        # emit error for unknown arguments
        # l += '/options:strict',
        l += cls._format_language_std(settings.lang_std)
        l += cls._format_language_conformance(settings.lang_conformance)
        l += cls._format_warning_settings(settings.warning_settings)
        l += ["-frtti"] if settings.enable_rtti else ["-fno-rtti"]
        l += ["-fcxx-exceptions"] if settings.enable_exceptions else ["-fno-cxx-exceptions"]
        l += cls._format_float_behavior(settings.float_behavior)
        l += ["-fno-complete-member-pointers"]
        # #TODO: unknown argument
        # l += ["-fms-runtime-lib=dll_dbg"] if settings.use_debug_runtime else ["-fms-runtime-lib=dll"]
        if settings.enable_debug_info:
            l += ["-g"]
        l += cls._format_optimization_mode(settings.optimization_mode)
        l += ["-mavx"]
        # if settings.enable_security_checks:
        #     # enable buffer security checks
        #     l += ['/GS']
        #     l += ['/sdl']
        #     # enable control flow guards
        #     l += ['/guard:cf']
        #     # enable EH continuation metadata(must be also present in linker args)
        #     l += ['/guard:ehcont']
        #     # enable all runtime checks.RTCc rejects conformant code, so it is not supported by
        #     # the C++ Standard Library
        #     l += ['/RTCsu']
        # else:
        #     # disable buffer security checks
        #     l += ['/GS-']
        #     # + '/sdl- ' // it overrides / GS -
        #     # disable control flow guards
        #     l += ['/guard:cf-']
        #     # disable EH continuation metadata
        #     l += ['/guard:ehcont-']
        return l

    def format_settings_args(self, settings: CompilerSettings) -> list[str]:
        return self._format_compiler_settings(settings)


class CompilerArgFormatterGCC(CompilerArgFormatterClang):
    pass


